import json
import os
import re
from urllib.parse import unquote

from requests import Session

from sgqlc.endpoint.requests import RequestsEndpoint
from sgqlc.operation import Operation

from .exceptions import (
    ForbiddenError, GraphQLError,
    NotAcceptableError, NotFoundError, UnauthorizedError
)
from .headquarters import Client
from .headquarters_schema import HeadquartersMutation


class HQBase(object):
    _apiprefix: str = ""

    def __init__(self, client: Client, workspace: str = None) -> None:
        self._hq = client
        self.workspace = workspace or client.workspace

    @property
    def url(self) -> str:
        if self.workspace:
            path = '/' + self.workspace + self._apiprefix
        else:
            path = self._apiprefix
        return self._hq.baseurl + path

    def _make_call(self, method: str, path: str, filepath: str = None, parser=None, use_login_session=False, **kwargs):
        if use_login_session:
            with Session() as login_session:
                response = login_session.request(method="post",
                                                 url=f"{self._hq.baseurl}/Account/LogOn",
                                                 data={"UserName": self._hq.session.auth[0],
                                                       "Password": self._hq.session.auth[1]})
                if response.status_code < 300:
                    response = login_session.request(method=method, url=path, **kwargs)
                else:
                    self._process_status_code(response)

        else:
            response = self._hq.session.request(method=method, url=path, **kwargs)
        if response.status_code < 300:
            if method == 'get':
                if 'application/json' in response.headers['Content-Type']:
                    if parser:
                        return parser(response.content)
                    else:
                        return json.loads(response.content)

                elif ('application/zip' in response.headers['Content-Type']
                      or 'application/octet-stream' in response.headers['Content-Type']):
                    return self._get_file_stream(filepath, response)
            else:
                return json.loads(response.content) if response.content else True

        else:
            self._process_status_code(response)

    def _call_mutation(self, method_name: str, fields: list = [], **kwargs):
        op = Operation(HeadquartersMutation)
        func = getattr(op, method_name)
        kwargs["workspace"] = self.workspace
        func(**kwargs).__fields__(*fields)
        cont = self._make_graphql_call(op)

        res = (op + cont)
        return getattr(res, method_name)

    def _make_graphql_call(self, op, **kwargs):
        if "session" not in kwargs:
            kwargs["session"] = self._hq.session
        endpoint = RequestsEndpoint(self._hq.baseurl + '/graphql', **kwargs)
        cont = endpoint(op)
        errors = cont.get('errors')
        if not errors:
            return cont

        try:
            rc = errors[0]['extensions']['code']
        except KeyError:
            rc = None
        if rc == 'AUTH_NOT_AUTHENTICATED':
            raise UnauthorizedError()
        else:
            raise GraphQLError(errors[0]['message'])

    @staticmethod
    def _process_status_code(response):
        rc = response.status_code
        if rc == 401:
            raise UnauthorizedError()
        elif rc == 403:
            raise ForbiddenError()
        elif rc == 404:
            raise NotFoundError(response.text)
        elif rc in [400, 406]:
            raise NotAcceptableError(response.text)
        else:
            response.raise_for_status()

    @staticmethod
    def _get_file_stream(filepath, response):
        d = response.headers['content-disposition']
        fname = re.findall(
            r"filename\*=utf-8''(.+)", d, flags=re.IGNORECASE)

        if not fname:
            fname = re.findall(
                r"filename[ ]*=([^;]+)", d, flags=re.IGNORECASE)

        fname = fname[0].strip().strip('"')
        outfile = os.path.join(filepath, unquote(fname))

        with open(outfile, 'wb') as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
        return outfile




from requests import Session

from .__about__ import __title__, __version__
from .models import Version


class Client(object):
    """Initializes the API client

    :param url: URL of the headquarters app
    :param api_user: API user name
    :param api_password: API user password
    :param workspace: Name of the workspace. If `None`, "primary" will be assumed
    """

    def __init__(self, url: str, api_user: str, api_password: str, workspace: str = "primary"):
        session = Session()
        session.auth = (api_user, api_password)
        signature = "python-{}/{}".format(__title__, __version__)
        session.headers.update({"User-Agent": signature})
        self.baseurl = url.rstrip("/")
        self.session = session
        self.workspace = workspace

    @property
    def version(self) -> Version:
        res = self.session.get("{}/.version".format(self.baseurl))
        if res.status_code == 200:
            return Version(res.text)



import datetime
import re
import sys
from enum import Enum
from typing import Dict, List, Optional, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, Extra, Field

from .headquarters_schema import Map
from .utils import get_properties, parse_date, to_camel, to_hex, to_qidentity

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


__all__ = ["Assignment", "ExportJob", "Map", ]


class QuestionType(Enum):
    SINGLE_SELECT = 0
    MULTI_SELECT = 3
    NUMERIC = 4
    DATE = 5
    GPS = 6
    TEXT = 7
    LIST = 9
    BARCODE = 10
    PICTURE = 11
    GEOGRAPHY = 12
    AUDIO = 13


class QuestionScope(Enum):
    INTERVIEWER = 0
    SUPERVISOR = 1
    HIDDEN = 3


class UserRole(Enum):
    ADMINISTRATOR = "Administrator"
    SUPERVISOR = "Supervisor"
    INTERVIEWER = "Interviewer"
    HEADQUARTERS = "Headquarter"
    OBSERVER = "Observer"
    APIUSER = "ApiUser"

    @property
    def code(self):
        if self.name == "ADMINISTRATOR":
            return 4
        elif self.name == "SUPERVISOR":
            return 2
        elif self.name == "INTERVIEWER":
            return 1
        elif self.name == "HEADQUARTERS":
            return 3
        elif self.name == "APIUSER":
            return 5
        else:
            return 0


class InterviewAction(Enum):
    SUPERVISOR_ASSIGNED = "SupervisorAssigned"
    INTERVIEWER_ASSIGNED = "InterviewerAssigned"
    FIRST_ANSWER_SET = "FirstAnswerSet"
    COMPLETED = "Completed"
    RESTARTED = "Restarted"
    APPROVED_BY_SUPERVISOR = "ApprovedBySupervisor"
    APPROVED_BY_HEADQUARTERS = "ApprovedByHeadquarter"
    REJECTED_BY_SUPERVISOR = "RejectedBySupervisor"
    REJECTED_BY_HEADQUARTERS = "RejectedByHeadquarter"
    DELETED = "Deleted"
    RESTORED = "Restored"
    UNAPPROVED_BY_HEADQUARTERS = "UnapprovedByHeadquarter"
    CREATED = "Created"
    INTERVIEW_RECEIVED_BY_TABLET = "InterviewReceivedByTablet"
    RESUMED = "Resumed"
    PAUSED = "Paused"
    TRANSLATION_SWITCHED = "TranslationSwitched"
    OPENED_BY_SUPERVISOR = "OpenedBySupervisor"
    CLOSED_BY_SUPERVISOR = "ClosedBySupervisor"

    @property
    def code(self):
        return self.value.ordinal()


class Assignment(object):
    def __init__(self, responsible: str, quantity: int, questionnaire_id,
                 identifying_data=None, email: str = "", password: str = "", webmode: bool = False,
                 audio_recording_enabled: bool = False, comments: str = "", protected_variables: list = None):
        """An assignment.

            :param responsible: responsible username
            :param quantity: [description]
            :param questionnaire_id: [description]
            :param email: [description], defaults to ""
            :param password: [description], defaults to ""
            :param webmode: [description], defaults to False
            :param audio_recording_enabled: [description], defaults to False
            :param comments: [description], defaults to ""
        """

        self.responsible = responsible
        self.quantity = quantity
        self.questionnaire_id = questionnaire_id
        self.identifying_data = identifying_data
        self.email = email
        self.password = password
        self.webmode = webmode
        self.audio_recording_enabled = audio_recording_enabled
        self.comments = comments
        self.protected_variables = protected_variables

    def __str__(self):
        return(str(self.__dict__))

    @classmethod
    def from_dict(cls, dict):
        obj = cls(
            responsible=dict["ResponsibleName"],
            quantity=dict["Quantity"],
            questionnaire_id=dict["QuestionnaireId"],
            identifying_data=dict["IdentifyingData"] if "IdentifyingData" in dict else [
            ],
            email=dict["Email"],
            password=dict["Password"],
            webmode=dict["WebMode"]
        )
        setattr(obj, "id", dict["Id"])
        setattr(obj, "responsible_id", dict["ResponsibleId"])
        setattr(obj, "interviews_count", dict["InterviewsCount"])
        setattr(obj, "archived", dict["Archived"])
        setattr(obj, "created_utc", dict["CreatedAtUtc"])
        setattr(obj, "updated_utc", dict["UpdatedAtUtc"])
        if "IsAudioRecordingEnabled" in dict:
            setattr(obj, "audio_recording_enabled",
                    dict["IsAudioRecordingEnabled"])
        return obj

    def to_json(self):
        return {
            "Responsible": self.responsible,
            "Quantity": self.quantity,
            "QuestionnaireId": self.questionnaire_id,
            "Email": self.email,
            "Password": self.password,
            "WebMode": self.webmode,
            "IsAudioRecordingEnabled": self.audio_recording_enabled,
            "Comments": self.comments,
            "IdentifyingData": self.identifying_data,
            "ProtectedVariables": self.protected_variables
        }


class ExportJob(object):
    def __init__(self,
                 questionnaire_identity: str,
                 export_type="Tabular",
                 interview_status="All",
                 from_date=None,
                 to_date=None,
                 access_token: str = None,
                 storage_type: str = None):
        """ExportJob object.

        :param questionnaire_identity: [description]
        :param export_type: [description], default to "Tabular"
        :param interview_status: [description], defaults to "All"
        :param from_date: [description], defaults to None
        :param to_date: [description], defaults to None
        :param access_token: [description], defaults to None
        :param storage_type: [description], defaults to None
        """

        if type(questionnaire_identity) is tuple:
            (questionnaire_id, questionnaire_version) = questionnaire_identity
            questionnaire_identity = to_qidentity(
                questionnaire_id, questionnaire_version)

        self.questionnaire_identity = questionnaire_identity
        self.export_type = export_type
        self.interview_status = interview_status
        self.from_date = from_date
        self.to_date = to_date
        self.access_token = access_token
        self.storage_type = storage_type
        self._export_status: str
        self._has_export_file: bool
        self._cancel_link: str
        self._download_link: str

    @property
    def export_status(self) -> str:
        return self._export_status

    @property
    def has_export_file(self) -> bool:
        return self._has_export_file

    @property
    def cancel_link(self) -> str:
        return self._cancel_link

    @property
    def download_link(self) -> str:
        return self._download_link

    def __str__(self) -> str:
        return(str(self.__dict__))

    @classmethod
    def from_dict(cls, dict):
        obj = cls(
            export_type=dict["ExportType"],
            questionnaire_identity=dict["QuestionnaireId"],
            interview_status=dict["InterviewStatus"],
            from_date=dict["From"] if "From" in dict else None,
            to_date=dict["To"] if "To" in dict else None,
            access_token=dict["AccessToken"] if "AccessToken" in dict else None,
            storage_type=dict["StorageType"] if "StorageType" in dict else None
        )
        setattr(obj, "job_id", dict["JobId"])
        setattr(obj, "_export_status", dict["ExportStatus"])
        setattr(obj, "start_date", parse_date(dict["StartDate"]))
        setattr(obj, "complete_date", parse_date(dict["CompleteDate"]))
        setattr(obj, "progress", dict["Progress"])
        if "ETA" in dict:
            setattr(obj, "eta", dict["ETA"])
        if "Links" in dict:
            if "Cancel" in dict["Links"]:
                setattr(obj, "_cancel_link", dict["Links"]["Cancel"])
            if "Download" in dict["Links"]:
                setattr(obj, "_download_link", dict["Links"]["Download"])
        setattr(obj, "_has_export_file", dict["HasExportFile"])
        return obj

    def to_json(self):
        ret = {
            "ExportType": self.export_type,
            "QuestionnaireId": self.questionnaire_identity,
            "InterviewStatus": self.interview_status,
            "From": self.from_date,
            "To": self.to_date,
            "AccessToken": self.access_token,
            "StorageType": self.storage_type,
        }
        return {k: v for k, v in ret.items() if v is not None}


class InterviewAnswers(object):
    def __init__(self, questionnaire_document=None):
        self._variables = {}
        self._public_ids = {}
        self._groups = {}
        self._data = {}
        self._raw_data = []
        if questionnaire_document:
            self._variables = get_properties(questionnaire_document)
            self._public_ids = {var.public_key.hex: variable_name for variable_name, var in self._variables.items()}
            self._groups = get_properties(questionnaire_document, groups=True, items=False)

    def __iter__(self):
        return iter(self._raw_data)

    def __str__(self) -> str:
        return(str(self._raw_data))

    def from_dict(self, answers: list):

        self._raw_data = answers
        for ans in self._raw_data:
            val = ans["Answer"]
            if val is None:
                continue
            id = to_hex(ans["QuestionId"]["Id"])
            variable_name = ans["VariableName"]
            if variable_name not in self._variables:
                self._variables[variable_name] = Question(variable_name=variable_name, public_key=id)
                self._public_ids[id] = variable_name
            key = (variable_name, ) + tuple(ans["QuestionId"]["RosterVector"])
            self._data[key] = val

    def dict(self, prefilling: bool = True) -> list:
        ret = []
        for t, answer in self._data.items():
            identity = self._variables[t[0]].public_key.hex
            if len(t) > 1:
                identity += "_" + "-".join(str(r) for r in t[1:])

            ans = answer
            q_type = self._variables[t[0]].question_type

            if q_type == QuestionType.LIST:
                ans_split = answer if type(answer) == list else answer.split("|")
                if prefilling:
                    ans = str(ans_split)
                else:
                    ans = [
                        {
                            "value": i + 1,
                            "isProtected": False,
                            "text": ans_split[i],
                        }
                        for i in range(len(ans_split))
                    ]

            elif q_type == QuestionType.MULTI_SELECT:
                ans_split = answer if type(answer) == list else answer.split(",")
                if prefilling:
                    ans = str(ans_split)
                else:
                    if self._variables[t[0]].yes_no_view:
                        ans = ans = [{"value": i, "isProtected": False, "yes": (i in ans_split)}
                                     for i in range(1, len(self._variables[t[0]].answers) + 1)]
                    else:
                        ans = ans_split

            elif q_type == QuestionType.SINGLE_SELECT and self._variables[t[0]].linked_to_roster_id and not prefilling:
                if type(ans) != list:
                    ans = [answer, ]
            ret.append({"Identity": identity, "Answer": ans})

        return ret

    def answer_variables(self) -> list:
        return [t[0] for t in self._data]

    def get_answer(self, variable: str = None, question_id: str = None, roster_vector: list = None):
        key = self._get_key(variable=variable, question_id=question_id, roster_vector=roster_vector)
        return self._data.get(key)

    def set_answer(self, answer, variable: str = None, question_id: str = None, roster_vector: list = []):
        key = self._get_key(variable=variable, question_id=question_id, roster_vector=roster_vector)
        variable = self._get_variable(variable=variable, question_id=question_id)
        self._data[key] = answer

    def remove_answer(self, variable: str = None, question_id: str = None, roster_vector: list = None):
        key = self._get_key(variable=variable, question_id=question_id, roster_vector=roster_vector)
        if key in self._data:
            del self._data[key]

    def needed_preloading(self) -> set:
        needed = set()
        for variable in self._data:
            triggers = self._find_roster_trigger(variable[0])
            if triggers:
                needed.update(triggers)
        return needed

    def _find_roster_trigger(self, variable) -> set:
        ret = set()
        parent_obj = self._groups.get(self._variables[variable].parent_id)
        if parent_obj.is_roster:
            trigger_id = parent_obj.roster_size_question_id.hex
            trigger_var = self._public_ids[trigger_id]
            if (trigger_var, ) not in self._data:
                ret.add(trigger_var)
            ret_in = self._find_roster_trigger(trigger_var)
            if ret_in:
                ret.update(ret_in)

        return ret

    def _get_key(self, roster_vector: list = [], **kwargs):
        key = (self._get_variable(**kwargs), )
        if roster_vector:
            key += tuple(roster_vector) if type(roster_vector) == list else tuple([roster_vector, ])
        return key

    def _get_variable(self, variable: str = None, question_id: str = None):
        if variable:
            if variable not in self._variables:
                raise TypeError("variable not found")
        else:
            if question_id:
                variable = self._public_ids.get(to_hex(question_id))
            else:
                raise TypeError(
                    "either 'variable' or 'question_id' argument is required")
        return variable


class BaseModelWithConfig(BaseModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True
        extra = Extra.allow


class AssignmentHistoryItemAdditionalData(BaseModelWithConfig):
    comment: str
    responsible: str = Field(alias="NewResponsible")


ASSIGNMENT_ACTION_TYPES = {
    "Unknown": 0,
    "Created": 1,
    "Archived": 2,
    "Deleted": 3,
    "ReceivedByTablet": 4,
    "UnArchived": 5,
    "AudioRecordingChanged": 6,
    "Reassigned": 7,
    "QuantityChanged": 8,
    "WebModeChanged": 9
}


class AssignmentHistoryItem(BaseModelWithConfig):
    action: str
    actor_name: str
    utc_date: datetime.datetime
    additional_data: AssignmentHistoryItemAdditionalData

    def to_datarow(self):
        return {
            "assignment__id": None,
            "date": self.utc_date.strftime("%Y-%m-%d"),
            "time": self.utc_date.strftime("%H:%M:%S"),
            "action": ASSIGNMENT_ACTION_TYPES[self.action],
            "originator": self.actor_name,
            "role": "",
            "responsible__name": self.additional_data.responsible,
            "responsible__role": "",
            "old__value": "",
            "new__value": "",
            "comment": self.additional_data.comment
        }


class ValidationCondition(BaseModelWithConfig):
    expression: str
    message: str
    severity: int


class Macro(BaseModelWithConfig):
    name: str
    content: str
    description: str = ""


class VariableType(Enum):
    BOOLEAN = 3
    DOUBLE = 2
    DATE = 4
    LONG = 1
    STRING = 5


class Variable(BaseModelWithConfig):
    obj_type: Literal["Variable"] = Field(alias="$type")
    public_key: UUID = Field(default_factory=uuid4)
    name: str
    label: str = ""
    type: VariableType
    expression: str
    do_not_export: bool
    variable_name: str


class Category(BaseModelWithConfig):
    id: UUID
    name: str


class LookupTable(BaseModelWithConfig):
    table_name: str
    file_name: str


class StaticText(BaseModelWithConfig):
    obj_type: Literal["StaticText"] = Field(alias="$type")
    public_key: UUID = Field(default_factory=uuid4)
    text: str
    attachment_name: str = ""
    hide_if_disabled: bool = False
    validation_conditions: List[ValidationCondition] = []
    variable_name: str


class QuestionProperties(BaseModelWithConfig):
    hide_instructions: bool = False
    use_formatting: bool = False


class Question(BaseModelWithConfig):
    condition_expression: str = ""
    featured: bool = False
    hide_if_disabled: bool = False
    instructions: str = ""
    properties: QuestionProperties = None
    public_key: UUID = Field(default_factory=uuid4)
    parent_id: UUID = None
    question_scope: QuestionScope = QuestionScope.INTERVIEWER
    question_text: str = ""
    question_type: QuestionType = QuestionType.TEXT
    stata_export_caption: str = ""
    validation_conditions: List[ValidationCondition] = []
    variable_label: Optional[str] = ""
    variable_name: str


class TextQuestion(Question):
    obj_type: Literal["TextQuestion"] = Field(alias="$type")
    mask: str = None
    value: str = ""


class NumericQuestion(Question):
    obj_type: Literal["NumericQuestion"] = Field(alias="$type",
                                                 default="NumericQuestion",
                                                 const="NumericQuestion")
    is_integer: bool = False
    value: float = None

    def __gt__(self, num):
        return self.value > num

    def __lt__(self, num):
        return self.value < num


class Answer(BaseModelWithConfig):
    answer_text: str
    answer_value: int


class SingleQuestion(Question):
    obj_type: Literal["SingleQuestion"] = Field(alias="$type",
                                                default="SingleQuestion",
                                                const="SingleQuestion")
    answers: List[Answer] = []
    cascade_from_question_id: UUID = None
    categories_id: UUID = None
    is_filtered_combobox: bool = False
    show_as_list: bool = False
    linked_to_question_id: UUID = None
    linked_to_roster_id: UUID = None


class MultiOptionsQuestion(Question):
    obj_type: Literal["MultyOptionsQuestion"] = Field(alias="$type",
                                                      default="MultyOptionsQuestion",
                                                      const="MultyOptionsQuestion")
    answers: List[Answer] = []
    are_answers_ordered: bool = False
    yes_no_view: bool = False
    linked_to_question_id: UUID = None
    linked_to_roster_id: UUID = None


class DateTimeQuestion(Question):
    obj_type: Literal["DateTimeQuestion"] = Field(alias="$type")
    is_timestamp: bool


class TextListQuestion(Question):
    obj_type: Literal["TextListQuestion"] = Field(alias="$type")
    max_answer_count: int


class GpsCoordinateQuestion(Question):
    obj_type: Literal["GpsCoordinateQuestion"] = Field(alias="$type",
                                                       default="GpsCoordinateQuestion",
                                                       const="GpsCoordinateQuestion")
    question_type: Literal[6]


class QRBarcodeQuestion(Question):
    obj_type: Literal["QRBarcodeQuestion"] = Field(alias="$type",
                                                   default="QRBarcodeQuestion",
                                                   const="QRBarcodeQuestion")


class MultimediaQuestion(Question):
    obj_type: Literal["MultimediaQuestion"] = Field(alias="$type",
                                                    default="MultimediaQuestion",
                                                    const="MultimediaQuestion")


class AudioQuestion(Question):
    obj_type: Literal["AudioQuestion"] = Field(alias="$type",
                                               default="AudioQuestion",
                                               const="AudioQuestion")


class AreaQuestion(Question):
    obj_type: Literal["AreaQuestion"] = Field(alias="$type",
                                              default="AreaQuestion",
                                              const="AreaQuestion")


class RosterSource(Enum):
    FIXED = 1
    OTHER = 0


class Group(BaseModelWithConfig):
    obj_type: Literal["Group"] = Field(alias="$type")
    description: str = ""
    display_mode: int = 0
    fixed_roster_titles: list = []
    is_flat_mode: bool = False
    is_plain_mode: bool = False
    is_roster: bool = False
    public_key: UUID = Field(default_factory=uuid4)
    parent_id: UUID = None
    roster_size_source: RosterSource = RosterSource.FIXED
    roster_size_question_id: UUID = None
    roster_title_question_id: UUID = None
    title: str
    variable_name: str
    children: List[Union[StaticText, NumericQuestion, TextQuestion, TextListQuestion,
                         SingleQuestion, MultiOptionsQuestion, DateTimeQuestion,
                         Variable, AreaQuestion, AudioQuestion, GpsCoordinateQuestion,
                         MultimediaQuestion, QRBarcodeQuestion, "Group"]] = []


class Attachment(BaseModelWithConfig):
    attachment_id: UUID
    content_id: str
    name: str


class QuestionnaireDocument(BaseModelWithConfig):
    attachments: List[Attachment]
    categories: List[Category]
    children: List[Group]
    creation_date: datetime.datetime
    lookup_tables: Dict[UUID, LookupTable]
    macros: Dict[UUID, Macro]
    public_key: UUID = Field(default_factory=uuid4)
    title: str
    translations: list
    variable_name: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def attach_parent_id(obj):
            if hasattr(obj, "children"):
                for ch in obj.children:
                    ch.parent_id = obj.public_key.hex
                    attach_parent_id(ch)

        attach_parent_id(self)


class InterviewerAction(BaseModelWithConfig):
    message: str
    time: datetime.datetime


class User(BaseModelWithConfig):
    user_name: str
    password: str
    role: UserRole = UserRole.INTERVIEWER
    supervisor: str = None
    full_name: str = None
    email: str = None
    phone_number: str = None


class Version():
    def __init__(self, version_string: str):
        pattern = r"(\d{1,2})\.(\d{1,2})(\.\d+)? \(build (\d+)\)"
        m = re.match(pattern, version_string, flags=re.IGNORECASE)
        if m:
            self.major = int(m.group(1))
            self.minor = int(m.group(2))
            self.patch = int(m.group(3)[1:]) if m.group(3) else 0
            self.build = int(m.group(4))
            self.version = version_string
        else:
            # dev version
            pattern = r"(\d{1,2})\.(\d{1,2})\.(\d+)\.(\d+)"
            m = re.match(pattern, version_string, flags=re.IGNORECASE)
            if m:
                self.major = int(m.group(1))
                self.minor = int(m.group(2))
                self.patch = int(m.group(3))
                self.build = int(m.group(4))
                self.version = version_string
            else:
                # somehow malformed version string, assume most recent
                self.version = version_string
                self.build = sys.maxsize

    def __repr__(self):
        return self.version

    def __lt__(self, other: "Version"):
        return self.build < other.build

    def __eq__(self, other: "Version"):
        return self.build == other.build


class AssignmentWebLink(BaseModelWithConfig):
    link: str = Field(alias="assignment__link")
    id: int = Field(alias="assignment__id")
    email: Optional[str] = Field(alias="assignment__email")
    password: Optional[str] = Field(alias="assignment__password")



from typing import Iterator

from sgqlc.operation import Operation

from .base import HQBase
from .headquarters_schema import (
    Assignment as GraphQLAssignment,
    AssignmentsFilter,
    CalendarEvent,
    HeadquartersQuery
)
from .models import Assignment, AssignmentHistoryItem
from .utils import filter_object, to_qidentity


class AssignmentsApi(HQBase):
    """ Set of functions to access and manipulate Assignments. """
    _apiprefix = "/api/v1/assignments"

    def get_list(self, questionnaire_id: str = None, questionnaire_version: int = None) -> Iterator[Assignment]:
        """Get list of assignments

        :param questionnaire_id: Filter by specific questionnaire id
        :param questionnaire_version: Filter by specific version number

        :returns: list of :class:`Assignment<ssaw.models.Assignment>` objects
        """
        path = self.url
        limit = 10
        offset = 1
        total_count = 11
        params = {
            'offset': offset,
            'limit': limit
        }
        if questionnaire_id and questionnaire_version:
            params['questionnaireId'] = to_qidentity(questionnaire_id, questionnaire_version)

        while offset < total_count:
            params['offset'] = offset
            r = self._make_call('get', path, params=params)
            total_count = r['TotalCount']
            offset += limit
            for item in r['Assignments']:
                yield Assignment.from_dict(item)

    def _get_list(self, fields: list = [], skip: int = None, take: int = None,
                  where: AssignmentsFilter = None, include_calendar_events: bool = False, **kwargs
                  ) -> Iterator[GraphQLAssignment]:

        args = {
            "workspace": self.workspace
        }
        if skip:
            args["skip"] = skip
        if take:
            args["take"] = take

        if where or kwargs:
            args['where'] = filter_object("AssignmentsFilter", where=where, **kwargs)

        op = Operation(HeadquartersQuery)
        q = op.assignments(**args)
        q.nodes.__fields__(*fields)
        if include_calendar_events:
            if type(include_calendar_events) in [list, tuple]:
                q.nodes.calendar_event.__fields__(*include_calendar_events)
            else:
                q.nodes.calendar_event.__fields__()
        cont = self._make_graphql_call(op)
        res = (op + cont).assignments

        yield from res.nodes

    def get_info(self, id: int) -> Assignment:
        """Get single assignment details

        :param id: Assignment Id

        :returns: Assignment object
        """
        path = self.url + "/{}".format(id)
        item = self._make_call("get", path)
        return Assignment.from_dict(item)

    def create(self, obj: Assignment) -> Assignment:
        """Create new assignment

        :param obj: Assignment object to be created

        :returns: Newly created Assignment object
        """
        path = self.url
        res = self._make_call("post", path, json=obj.to_json())
        return Assignment.from_dict(res['Assignment'])

    def archive(self, id: int) -> None:
        """Archive assignment

        :param id: Assignment Id
        """
        path = self.url + "/{}/archive".format(id)
        self._make_call("patch", path)

    def unarchive(self, id: int) -> None:
        """Unarchive assignment

        :param id: Assignment Id
        """
        path = self.url + "/{}/unarchive".format(id)
        self._make_call("patch", path)

    def assign(self, id: int, responsible: str) -> Assignment:
        """Assign new responsible person for assignment

        :param id: Assignment Id
        :param responsible: Username of the new responsible

        :returns: Modified Assignment object
        """
        path = self.url + "/{}/assign".format(id)
        res = self._make_call("patch", path, json={"Responsible": responsible})
        return Assignment.from_dict(res)

    def get_quantity_settings(self, id: int) -> bool:
        """Checi if quantity may be edited for the assignment

        :param id: Assignment Id

        :returns: `True` if quantity can be edited, `False` otherwise
        """
        path = self.url + "/{}/assignmentQuantitySettings".format(id)
        res = self._make_call("get", path)
        return res['CanChangeQuantity']

    def update_quantity(self, id: int, quantity: int) -> Assignment:
        """Change maximum quantity of interviews to be created

        :param id: Assignment Id
        :param quantity: new quantity of interviews to be collected

        :returns: Modified Assignment object

        """
        if not isinstance(quantity, int):
            raise TypeError('quantity must be a number')
        path = self.url + "/{}/changequantity".format(id)
        headers = {"Content-Type": "application/json-patch+json"}
        res = self._make_call("patch", path, data=str(quantity), headers=headers)
        return Assignment.from_dict(res)

    def close(self, id: int) -> None:
        """Close assignment by setting Size to the number of collected interviews

        :param id: Assignment Id
        """
        path = self.url + "/{}/close".format(id)
        self._make_call("post", path)

    def get_history(self, id: int) -> Iterator[dict]:
        page_size = 10
        start = 0
        total_count = 11
        params = {
            'start': start,
            'length': page_size
        }
        path = self.url + "/{}/history".format(id)
        while start < total_count:
            params['start'] = start
            r = self._make_call('get', path, params=params)
            total_count = r['RecordsFiltered']
            start += page_size
            for item in r['History']:
                print(item)
                yield AssignmentHistoryItem(**item)

    def get_recordaudio(self, id: int) -> bool:
        """Get status of audio recording for the assignment

        :param id: Assignment Id

        :returns: `True` if audio recording is enabled, `False` otherwise
        """
        path = self.url + "/{}/recordAudio".format(id)
        res = self._make_call("get", path)
        return res['Enabled']

    def update_recordaudio(self, id: int, enabled: bool) -> None:
        """Turn recording of audio for the assignment

        :param id: Assignment Id
        :param enabled: `True` to turn audio recording on, `False` to turn it off
        """
        if not isinstance(enabled, bool):
            raise TypeError('enabled must be either True or False')
        path = self.url + "/{}/recordAudio".format(id)
        self._make_call("patch", path, json={'Enabled': enabled})

    def get_calendar_event(self, id: int) -> CalendarEvent:
        """Get calendar event associated with the assignment
        """
        assignment = self._get_assignment_by_id(fields=["id"], id=id, include_calendar_events=True)

        return assignment.calendar_event

    def set_calendar_event(self, id: int, new_start: str,
                           start_timezone: str, comment: str = None) -> CalendarEvent:
        """Add new calendar event to the assignment, or update the existing one

        :param id: assignment id to update
        :param new_start: start date of the calendar event
        :param start_timezone: timezone string for the start date
        :param comment: add comment to the calendar event

        :returns: :class:`CalendarEvent' object
        """
        assignment = self._get_assignment_by_id(fields=["id"], id=id, include_calendar_events=["public_key"])

        kwargs = {
            "new_start": new_start,
            "start_timezone": start_timezone,
            "comment": comment
        }
        if hasattr(assignment.calendar_event, "public_key"):
            kwargs["method_name"] = "update_calendar_event"
            kwargs["public_key"] = assignment.calendar_event.public_key
        else:
            kwargs["method_name"] = "add_assignment_calendar_event"
            kwargs["assignment_id"] = assignment.id

        return self._call_mutation(**kwargs)

    def delete_calendar_event(self, id: int) -> CalendarEvent:
        """Remove calendar event associated with the assignment
        """
        assignment = self._get_assignment_by_id(fields=["id"], id=id, include_calendar_events=["public_key"])

        if hasattr(assignment.calendar_event, "public_key"):
            return self._call_mutation(method_name="delete_calendar_event",
                                       public_key=assignment.calendar_event.public_key)

    def _get_assignment_by_id(self, **kwargs):
        # expecting fields, id, include_calendar_events
        try:
            return next(self._get_list(**kwargs))
        except StopIteration:
            raise ValueError("assignment was not found")



import sgqlc.types
import sgqlc.types.datetime


headquarters_schema = sgqlc.types.Schema()



########################################################################
# Scalars and Enumerations
########################################################################
class Any(sgqlc.types.Scalar):
    __schema__ = headquarters_schema


class ApplyPolicy(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('BEFORE_RESOLVER', 'AFTER_RESOLVER')


Boolean = sgqlc.types.Boolean

Date = sgqlc.types.datetime.Date

DateTime = sgqlc.types.datetime.DateTime

class Decimal(sgqlc.types.Scalar):
    __schema__ = headquarters_schema


class EntityType(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('SECTION', 'QUESTION', 'STATICTEXT', 'VARIABLE', 'ROSTER')


Float = sgqlc.types.Float

class GeoJSONObjectType(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('Point', 'MultiPoint', 'LineString', 'MultiLineString', 'Polygon', 'MultiPolygon', 'GeometryCollection', 'Feature', 'FeatureCollection')


ID = sgqlc.types.ID

Int = sgqlc.types.Int

class InterviewActionFlags(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('CANBEREASSIGNED', 'CANBEDELETED', 'CANBEAPPROVED', 'CANBEUNAPPROVEDBYHQ', 'CANBEREJECTED', 'CANBERESTARTED', 'CANBEOPENED', 'CANCHANGETOCAPI', 'CANCHANGETOCAWI')


class InterviewMode(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('UNKNOWN', 'CAPI', 'CAWI')


class InterviewStatus(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('RESTORED', 'CREATED', 'SUPERVISORASSIGNED', 'INTERVIEWERASSIGNED', 'REJECTEDBYSUPERVISOR', 'READYFORINTERVIEW', 'SENTTOCAPI', 'RESTARTED', 'COMPLETED', 'APPROVEDBYSUPERVISOR', 'REJECTEDBYHEADQUARTERS', 'APPROVEDBYHEADQUARTERS', 'DELETED')


class Long(sgqlc.types.Scalar):
    __schema__ = headquarters_schema


class QuestionScope(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('INTERVIEWER', 'SUPERVISOR', 'HEADQUARTER', 'HIDDEN')


class QuestionType(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('SINGLEOPTION', 'MULTYOPTION', 'NUMERIC', 'DATETIME', 'GPSCOORDINATES', 'TEXT', 'TEXTLIST', 'QRBARCODE', 'MULTIMEDIA', 'AREA', 'AUDIO')


class SortEnumType(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('ASC', 'DESC')


String = sgqlc.types.String

class UserRoles(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('ADMINISTRATOR', 'SUPERVISOR', 'INTERVIEWER', 'HEADQUARTER', 'OBSERVER', 'APIUSER')


class Uuid(sgqlc.types.Scalar):
    __schema__ = headquarters_schema


class VariableType(sgqlc.types.Enum):
    __schema__ = headquarters_schema
    __choices__ = ('LONGINTEGER', 'DOUBLE', 'BOOLEAN', 'DATETIME', 'STRING')



########################################################################
# Input Objects
########################################################################
class AssignmentsFilter(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'id', 'questionnaire_id', 'archived', 'responsible_id', 'web_mode', 'received_by_tablet_at_utc')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('AssignmentsFilter')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('AssignmentsFilter')), graphql_name='or')
    id = sgqlc.types.Field('ComparableInt32OperationFilterInput', graphql_name='id')
    questionnaire_id = sgqlc.types.Field('QuestionnaireIdentity', graphql_name='questionnaireId')
    archived = sgqlc.types.Field('BooleanOperationFilterInput', graphql_name='archived')
    responsible_id = sgqlc.types.Field('ComparableGuidOperationFilterInput', graphql_name='responsibleId')
    web_mode = sgqlc.types.Field('BooleanOperationFilterInput', graphql_name='webMode')
    received_by_tablet_at_utc = sgqlc.types.Field('ComparableNullableOfDateTimeOperationFilterInput', graphql_name='receivedByTabletAtUtc')


class BooleanOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq')
    eq = sgqlc.types.Field(Boolean, graphql_name='eq')
    neq = sgqlc.types.Field(Boolean, graphql_name='neq')


class ComparableDateTimeOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(DateTime, graphql_name='eq')
    neq = sgqlc.types.Field(DateTime, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(DateTime)), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(DateTime)), graphql_name='nin')
    gt = sgqlc.types.Field(DateTime, graphql_name='gt')
    ngt = sgqlc.types.Field(DateTime, graphql_name='ngt')
    gte = sgqlc.types.Field(DateTime, graphql_name='gte')
    ngte = sgqlc.types.Field(DateTime, graphql_name='ngte')
    lt = sgqlc.types.Field(DateTime, graphql_name='lt')
    nlt = sgqlc.types.Field(DateTime, graphql_name='nlt')
    lte = sgqlc.types.Field(DateTime, graphql_name='lte')
    nlte = sgqlc.types.Field(DateTime, graphql_name='nlte')


class ComparableGuidOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(Uuid, graphql_name='eq')
    neq = sgqlc.types.Field(Uuid, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(Uuid)), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(Uuid)), graphql_name='nin')
    gt = sgqlc.types.Field(Uuid, graphql_name='gt')
    ngt = sgqlc.types.Field(Uuid, graphql_name='ngt')
    gte = sgqlc.types.Field(Uuid, graphql_name='gte')
    ngte = sgqlc.types.Field(Uuid, graphql_name='ngte')
    lt = sgqlc.types.Field(Uuid, graphql_name='lt')
    nlt = sgqlc.types.Field(Uuid, graphql_name='nlt')
    lte = sgqlc.types.Field(Uuid, graphql_name='lte')
    nlte = sgqlc.types.Field(Uuid, graphql_name='nlte')


class ComparableInt32OperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(Int, graphql_name='eq')
    neq = sgqlc.types.Field(Int, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(Int)), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(Int)), graphql_name='nin')
    gt = sgqlc.types.Field(Int, graphql_name='gt')
    ngt = sgqlc.types.Field(Int, graphql_name='ngt')
    gte = sgqlc.types.Field(Int, graphql_name='gte')
    ngte = sgqlc.types.Field(Int, graphql_name='ngte')
    lt = sgqlc.types.Field(Int, graphql_name='lt')
    nlt = sgqlc.types.Field(Int, graphql_name='nlt')
    lte = sgqlc.types.Field(Int, graphql_name='lte')
    nlte = sgqlc.types.Field(Int, graphql_name='nlte')


class ComparableInt64OperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(Long, graphql_name='eq')
    neq = sgqlc.types.Field(Long, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(Long)), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(Long)), graphql_name='nin')
    gt = sgqlc.types.Field(Long, graphql_name='gt')
    ngt = sgqlc.types.Field(Long, graphql_name='ngt')
    gte = sgqlc.types.Field(Long, graphql_name='gte')
    ngte = sgqlc.types.Field(Long, graphql_name='ngte')
    lt = sgqlc.types.Field(Long, graphql_name='lt')
    nlt = sgqlc.types.Field(Long, graphql_name='nlt')
    lte = sgqlc.types.Field(Long, graphql_name='lte')
    nlte = sgqlc.types.Field(Long, graphql_name='nlte')


class ComparableNullableOfDateTimeOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(DateTime, graphql_name='eq')
    neq = sgqlc.types.Field(DateTime, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(DateTime), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(DateTime), graphql_name='nin')
    gt = sgqlc.types.Field(DateTime, graphql_name='gt')
    ngt = sgqlc.types.Field(DateTime, graphql_name='ngt')
    gte = sgqlc.types.Field(DateTime, graphql_name='gte')
    ngte = sgqlc.types.Field(DateTime, graphql_name='ngte')
    lt = sgqlc.types.Field(DateTime, graphql_name='lt')
    nlt = sgqlc.types.Field(DateTime, graphql_name='nlt')
    lte = sgqlc.types.Field(DateTime, graphql_name='lte')
    nlte = sgqlc.types.Field(DateTime, graphql_name='nlte')


class ComparableNullableOfDecimalOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(Decimal, graphql_name='eq')
    neq = sgqlc.types.Field(Decimal, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(Decimal), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(Decimal), graphql_name='nin')
    gt = sgqlc.types.Field(Decimal, graphql_name='gt')
    ngt = sgqlc.types.Field(Decimal, graphql_name='ngt')
    gte = sgqlc.types.Field(Decimal, graphql_name='gte')
    ngte = sgqlc.types.Field(Decimal, graphql_name='ngte')
    lt = sgqlc.types.Field(Decimal, graphql_name='lt')
    nlt = sgqlc.types.Field(Decimal, graphql_name='nlt')
    lte = sgqlc.types.Field(Decimal, graphql_name='lte')
    nlte = sgqlc.types.Field(Decimal, graphql_name='nlte')


class ComparableNullableOfDoubleOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(Float, graphql_name='eq')
    neq = sgqlc.types.Field(Float, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(Float), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(Float), graphql_name='nin')
    gt = sgqlc.types.Field(Float, graphql_name='gt')
    ngt = sgqlc.types.Field(Float, graphql_name='ngt')
    gte = sgqlc.types.Field(Float, graphql_name='gte')
    ngte = sgqlc.types.Field(Float, graphql_name='ngte')
    lt = sgqlc.types.Field(Float, graphql_name='lt')
    nlt = sgqlc.types.Field(Float, graphql_name='nlt')
    lte = sgqlc.types.Field(Float, graphql_name='lte')
    nlte = sgqlc.types.Field(Float, graphql_name='nlte')


class ComparableNullableOfInt32OperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(Int, graphql_name='eq')
    neq = sgqlc.types.Field(Int, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(Int), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(Int), graphql_name='nin')
    gt = sgqlc.types.Field(Int, graphql_name='gt')
    ngt = sgqlc.types.Field(Int, graphql_name='ngt')
    gte = sgqlc.types.Field(Int, graphql_name='gte')
    ngte = sgqlc.types.Field(Int, graphql_name='ngte')
    lt = sgqlc.types.Field(Int, graphql_name='lt')
    nlt = sgqlc.types.Field(Int, graphql_name='nlt')
    lte = sgqlc.types.Field(Int, graphql_name='lte')
    nlte = sgqlc.types.Field(Int, graphql_name='nlte')


class ComparableNullableOfInt64OperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin', 'gt', 'ngt', 'gte', 'ngte', 'lt', 'nlt', 'lte', 'nlte')
    eq = sgqlc.types.Field(Long, graphql_name='eq')
    neq = sgqlc.types.Field(Long, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(Long), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(Long), graphql_name='nin')
    gt = sgqlc.types.Field(Long, graphql_name='gt')
    ngt = sgqlc.types.Field(Long, graphql_name='ngt')
    gte = sgqlc.types.Field(Long, graphql_name='gte')
    ngte = sgqlc.types.Field(Long, graphql_name='ngte')
    lt = sgqlc.types.Field(Long, graphql_name='lt')
    nlt = sgqlc.types.Field(Long, graphql_name='nlt')
    lte = sgqlc.types.Field(Long, graphql_name='lte')
    nlte = sgqlc.types.Field(Long, graphql_name='nlte')


class IdentifyEntityValueFilter(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'answer_code', 'value', 'value_lower_case', 'entity', 'value_bool', 'value_date', 'value_double', 'value_long', 'is_enabled')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('IdentifyEntityValueFilter')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('IdentifyEntityValueFilter')), graphql_name='or')
    answer_code = sgqlc.types.Field(ComparableNullableOfDecimalOperationFilterInput, graphql_name='answerCode')
    value = sgqlc.types.Field('StringOperationFilterInput', graphql_name='value')
    value_lower_case = sgqlc.types.Field('StringOperationFilterInput', graphql_name='valueLowerCase')
    entity = sgqlc.types.Field('QuestionnaireItemsFilter', graphql_name='entity')
    value_bool = sgqlc.types.Field(BooleanOperationFilterInput, graphql_name='valueBool')
    value_date = sgqlc.types.Field(ComparableNullableOfDateTimeOperationFilterInput, graphql_name='valueDate')
    value_double = sgqlc.types.Field(ComparableNullableOfDoubleOperationFilterInput, graphql_name='valueDouble')
    value_long = sgqlc.types.Field(ComparableNullableOfInt64OperationFilterInput, graphql_name='valueLong')
    is_enabled = sgqlc.types.Field(BooleanOperationFilterInput, graphql_name='isEnabled')


class InterviewModeOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin')
    eq = sgqlc.types.Field(InterviewMode, graphql_name='eq')
    neq = sgqlc.types.Field(InterviewMode, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(InterviewMode)), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(InterviewMode)), graphql_name='nin')


class InterviewSort(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('key', 'created_date', 'update_date_utc', 'responsible_name', 'responsible_role', 'assignment_id', 'errors_count', 'status', 'received_by_interviewer_at_utc', 'questionnaire_id', 'questionnaire_version', 'id', 'not_answered_count')
    key = sgqlc.types.Field(SortEnumType, graphql_name='key')
    created_date = sgqlc.types.Field(SortEnumType, graphql_name='createdDate')
    update_date_utc = sgqlc.types.Field(SortEnumType, graphql_name='updateDateUtc')
    responsible_name = sgqlc.types.Field(SortEnumType, graphql_name='responsibleName')
    responsible_role = sgqlc.types.Field(SortEnumType, graphql_name='responsibleRole')
    assignment_id = sgqlc.types.Field(SortEnumType, graphql_name='assignmentId')
    errors_count = sgqlc.types.Field(SortEnumType, graphql_name='errorsCount')
    status = sgqlc.types.Field(SortEnumType, graphql_name='status')
    received_by_interviewer_at_utc = sgqlc.types.Field(SortEnumType, graphql_name='receivedByInterviewerAtUtc')
    questionnaire_id = sgqlc.types.Field(SortEnumType, graphql_name='questionnaireId')
    questionnaire_version = sgqlc.types.Field(SortEnumType, graphql_name='questionnaireVersion')
    id = sgqlc.types.Field(SortEnumType, graphql_name='id')
    not_answered_count = sgqlc.types.Field(SortEnumType, graphql_name='notAnsweredCount')


class InterviewStatusOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin')
    eq = sgqlc.types.Field(InterviewStatus, graphql_name='eq')
    neq = sgqlc.types.Field(InterviewStatus, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(InterviewStatus)), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(InterviewStatus)), graphql_name='nin')


class InterviewsFilter(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'status', 'interview_mode', 'questionnaire_id', 'questionnaire_variable', 'questionnaire_version', 'key', 'not_answered_count', 'client_key', 'assignment_id', 'created_date', 'responsible_name', 'supervisor_name', 'responsible_role', 'update_date_utc', 'received_by_interviewer_at_utc', 'errors_count', 'identifying_data')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('InterviewsFilter')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('InterviewsFilter')), graphql_name='or')
    status = sgqlc.types.Field(InterviewStatusOperationFilterInput, graphql_name='status')
    interview_mode = sgqlc.types.Field(InterviewModeOperationFilterInput, graphql_name='interviewMode')
    questionnaire_id = sgqlc.types.Field(ComparableGuidOperationFilterInput, graphql_name='questionnaireId')
    questionnaire_variable = sgqlc.types.Field('StringOperationFilterInput', graphql_name='questionnaireVariable')
    questionnaire_version = sgqlc.types.Field(ComparableInt64OperationFilterInput, graphql_name='questionnaireVersion')
    key = sgqlc.types.Field('StringOperationFilterInput', graphql_name='key')
    not_answered_count = sgqlc.types.Field(ComparableNullableOfInt32OperationFilterInput, graphql_name='notAnsweredCount')
    client_key = sgqlc.types.Field('StringOperationFilterInput', graphql_name='clientKey')
    assignment_id = sgqlc.types.Field(ComparableNullableOfInt32OperationFilterInput, graphql_name='assignmentId')
    created_date = sgqlc.types.Field(ComparableDateTimeOperationFilterInput, graphql_name='createdDate')
    responsible_name = sgqlc.types.Field('StringOperationFilterInput', graphql_name='responsibleNameLowerCase')
    supervisor_name = sgqlc.types.Field('StringOperationFilterInput', graphql_name='supervisorNameLowerCase')
    responsible_role = sgqlc.types.Field('UserRolesOperationFilterInput', graphql_name='responsibleRole')
    update_date_utc = sgqlc.types.Field(ComparableDateTimeOperationFilterInput, graphql_name='updateDateUtc')
    received_by_interviewer_at_utc = sgqlc.types.Field(ComparableNullableOfDateTimeOperationFilterInput, graphql_name='receivedByInterviewerAtUtc')
    errors_count = sgqlc.types.Field(ComparableInt32OperationFilterInput, graphql_name='errorsCount')
    identifying_data = sgqlc.types.Field('ListFilterInputTypeOfIdentifyEntityValueFilterInput', graphql_name='identifyingData')


class ListFilterInputTypeOfIdentifyEntityValueFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('all', 'none', 'some', 'any')
    all = sgqlc.types.Field(IdentifyEntityValueFilter, graphql_name='all')
    none = sgqlc.types.Field(IdentifyEntityValueFilter, graphql_name='none')
    some = sgqlc.types.Field(IdentifyEntityValueFilter, graphql_name='some')
    any = sgqlc.types.Field(Boolean, graphql_name='any')


class ListFilterInputTypeOfUserMapFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('all', 'none', 'some', 'any')
    all = sgqlc.types.Field('UserMapFilterInput', graphql_name='all')
    none = sgqlc.types.Field('UserMapFilterInput', graphql_name='none')
    some = sgqlc.types.Field('UserMapFilterInput', graphql_name='some')
    any = sgqlc.types.Field(Boolean, graphql_name='any')


class MapReportFilter(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'interview_filter')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('MapReportFilter')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('MapReportFilter')), graphql_name='or')
    interview_filter = sgqlc.types.Field(InterviewsFilter, graphql_name='interviewFilter')


class MapsFilter(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'file_name', 'import_date_utc', 'size', 'users')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('MapsFilter')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('MapsFilter')), graphql_name='or')
    file_name = sgqlc.types.Field('StringOperationFilterInput', graphql_name='fileName')
    import_date_utc = sgqlc.types.Field(ComparableNullableOfDateTimeOperationFilterInput, graphql_name='importDateUtc')
    size = sgqlc.types.Field(ComparableInt64OperationFilterInput, graphql_name='size')
    users = sgqlc.types.Field(ListFilterInputTypeOfUserMapFilterInput, graphql_name='users')


class MapsSort(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('file_name', 'import_date_utc', 'size')
    file_name = sgqlc.types.Field(SortEnumType, graphql_name='fileName')
    import_date_utc = sgqlc.types.Field(SortEnumType, graphql_name='importDateUtc')
    size = sgqlc.types.Field(SortEnumType, graphql_name='size')


class NullableOfQuestionScopeOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin')
    eq = sgqlc.types.Field(QuestionScope, graphql_name='eq')
    neq = sgqlc.types.Field(QuestionScope, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(QuestionScope), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(QuestionScope), graphql_name='nin')


class QuestionFilter(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'variable', 'scope', 'identifying')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('QuestionFilter')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('QuestionFilter')), graphql_name='or')
    variable = sgqlc.types.Field('StringOperationFilterInput', graphql_name='variable')
    scope = sgqlc.types.Field(NullableOfQuestionScopeOperationFilterInput, graphql_name='scope')
    identifying = sgqlc.types.Field(BooleanOperationFilterInput, graphql_name='identifying')


class QuestionnaireIdentity(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'id', 'version')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('QuestionnaireIdentity')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('QuestionnaireIdentity')), graphql_name='or')
    id = sgqlc.types.Field(ComparableGuidOperationFilterInput, graphql_name='id')
    version = sgqlc.types.Field(ComparableInt64OperationFilterInput, graphql_name='version')


class QuestionnaireItemsFilter(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'title', 'variable', 'scope', 'identifying', 'included_in_reporting_at_utc')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('QuestionnaireItemsFilter')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('QuestionnaireItemsFilter')), graphql_name='or')
    title = sgqlc.types.Field('StringOperationFilterInput', graphql_name='title')
    variable = sgqlc.types.Field('StringOperationFilterInput', graphql_name='variable')
    scope = sgqlc.types.Field(NullableOfQuestionScopeOperationFilterInput, graphql_name='scope')
    identifying = sgqlc.types.Field(BooleanOperationFilterInput, graphql_name='identifying')
    included_in_reporting_at_utc = sgqlc.types.Field(ComparableNullableOfDateTimeOperationFilterInput, graphql_name='includedInReportingAtUtc')


class RoleFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'eq')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('RoleFilterInput')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('RoleFilterInput')), graphql_name='or')
    eq = sgqlc.types.Field(UserRoles, graphql_name='eq')


class StringOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'eq', 'neq', 'contains', 'ncontains', 'in_', 'nin', 'starts_with', 'nstarts_with', 'ends_with', 'nends_with')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('StringOperationFilterInput')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('StringOperationFilterInput')), graphql_name='or')
    eq = sgqlc.types.Field(String, graphql_name='eq')
    neq = sgqlc.types.Field(String, graphql_name='neq')
    contains = sgqlc.types.Field(String, graphql_name='contains')
    ncontains = sgqlc.types.Field(String, graphql_name='ncontains')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='nin')
    starts_with = sgqlc.types.Field(String, graphql_name='startsWith')
    nstarts_with = sgqlc.types.Field(String, graphql_name='nstartsWith')
    ends_with = sgqlc.types.Field(String, graphql_name='endsWith')
    nends_with = sgqlc.types.Field(String, graphql_name='nendsWith')


class StringOperationFilterInputType(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'eq', 'neq', 'contains', 'ncontains', 'in_', 'nin', 'starts_with', 'nstarts_with')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('StringOperationFilterInputType')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('StringOperationFilterInputType')), graphql_name='or')
    eq = sgqlc.types.Field(String, graphql_name='eq')
    neq = sgqlc.types.Field(String, graphql_name='neq')
    contains = sgqlc.types.Field(String, graphql_name='contains')
    ncontains = sgqlc.types.Field(String, graphql_name='ncontains')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='nin')
    starts_with = sgqlc.types.Field(String, graphql_name='startsWith')
    nstarts_with = sgqlc.types.Field(String, graphql_name='nstartsWith')


class UserMapFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'id', 'user_name', 'map')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('UserMapFilterInput')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('UserMapFilterInput')), graphql_name='or')
    id = sgqlc.types.Field(ComparableInt32OperationFilterInput, graphql_name='id')
    user_name = sgqlc.types.Field(StringOperationFilterInput, graphql_name='userName')
    map = sgqlc.types.Field(MapsFilter, graphql_name='map')


class UserRolesOperationFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('eq', 'neq', 'in_', 'nin')
    eq = sgqlc.types.Field(UserRoles, graphql_name='eq')
    neq = sgqlc.types.Field(UserRoles, graphql_name='neq')
    in_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(UserRoles)), graphql_name='in')
    nin = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null(UserRoles)), graphql_name='nin')


class UsersFilterInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('and_', 'or_', 'user_name', 'full_name', 'is_archived', 'creation_date', 'email', 'phone_number', 'id', 'role')
    and_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('UsersFilterInput')), graphql_name='and')
    or_ = sgqlc.types.Field(sgqlc.types.list_of(sgqlc.types.non_null('UsersFilterInput')), graphql_name='or')
    user_name = sgqlc.types.Field(StringOperationFilterInput, graphql_name='userName')
    full_name = sgqlc.types.Field(StringOperationFilterInput, graphql_name='fullName')
    is_archived = sgqlc.types.Field(BooleanOperationFilterInput, graphql_name='isArchived')
    creation_date = sgqlc.types.Field(ComparableDateTimeOperationFilterInput, graphql_name='creationDate')
    email = sgqlc.types.Field(StringOperationFilterInput, graphql_name='email')
    phone_number = sgqlc.types.Field(StringOperationFilterInput, graphql_name='phoneNumber')
    id = sgqlc.types.Field(ComparableGuidOperationFilterInput, graphql_name='id')
    role = sgqlc.types.Field(RoleFilterInput, graphql_name='role')


class UsersSortInput(sgqlc.types.Input):
    __schema__ = headquarters_schema
    __field_names__ = ('user_name', 'creation_date', 'full_name')
    user_name = sgqlc.types.Field(SortEnumType, graphql_name='userName')
    creation_date = sgqlc.types.Field(SortEnumType, graphql_name='creationDate')
    full_name = sgqlc.types.Field(SortEnumType, graphql_name='fullName')



########################################################################
# Output Objects and Interfaces
########################################################################
class Assignment(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('archived', 'created_at_utc', 'email', 'id', 'interviews_needed', 'received_by_tablet_at_utc', 'responsible_id', 'web_mode', 'calendar_event')
    archived = sgqlc.types.Field(sgqlc.types.non_null(Boolean), graphql_name='archived')
    created_at_utc = sgqlc.types.Field(sgqlc.types.non_null(DateTime), graphql_name='createdAtUtc')
    email = sgqlc.types.Field(String, graphql_name='email')
    id = sgqlc.types.Field(sgqlc.types.non_null(ID), graphql_name='id')
    interviews_needed = sgqlc.types.Field(Int, graphql_name='interviewsNeeded')
    received_by_tablet_at_utc = sgqlc.types.Field(DateTime, graphql_name='receivedByTabletAtUtc')
    responsible_id = sgqlc.types.Field(sgqlc.types.non_null(Uuid), graphql_name='responsibleId')
    web_mode = sgqlc.types.Field(sgqlc.types.non_null(Boolean), graphql_name='webMode')
    calendar_event = sgqlc.types.Field('CalendarEvent', graphql_name='calendarEvent')


class CalendarEvent(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('assignment_id', 'comment', 'creator_user_id', 'interview_id', 'interview_key', 'is_completed', 'public_key', 'start_timezone', 'start_utc', 'update_date_utc')
    assignment_id = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='assignmentId')
    comment = sgqlc.types.Field(String, graphql_name='comment')
    creator_user_id = sgqlc.types.Field(sgqlc.types.non_null(Uuid), graphql_name='creatorUserId')
    interview_id = sgqlc.types.Field(Uuid, graphql_name='interviewId')
    interview_key = sgqlc.types.Field(String, graphql_name='interviewKey')
    is_completed = sgqlc.types.Field(sgqlc.types.non_null(Boolean), graphql_name='isCompleted')
    public_key = sgqlc.types.Field(sgqlc.types.non_null(Uuid), graphql_name='publicKey')
    start_timezone = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='startTimezone')
    start_utc = sgqlc.types.Field(sgqlc.types.non_null(DateTime), graphql_name='startUtc')
    update_date_utc = sgqlc.types.Field(sgqlc.types.non_null(DateTime), graphql_name='updateDateUtc')


class CategoricalOption(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('parent_value', 'title', 'value')
    parent_value = sgqlc.types.Field(Int, graphql_name='parentValue')
    title = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='title')
    value = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='value')


class Entity(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('identifying', 'label', 'options', 'question_text', 'scope', 'type', 'variable', 'variable_type')
    identifying = sgqlc.types.Field(Boolean, graphql_name='identifying')
    label = sgqlc.types.Field(String, graphql_name='label')
    options = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(CategoricalOption))), graphql_name='options')
    question_text = sgqlc.types.Field(String, graphql_name='questionText')
    scope = sgqlc.types.Field(QuestionScope, graphql_name='scope')
    type = sgqlc.types.Field(sgqlc.types.non_null(QuestionType), graphql_name='type')
    variable = sgqlc.types.Field(String, graphql_name='variable')
    variable_type = sgqlc.types.Field(sgqlc.types.non_null(VariableType), graphql_name='variableType')


class Feature(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('type', 'id', 'geometry', 'properties')
    type = sgqlc.types.Field(sgqlc.types.non_null(GeoJSONObjectType), graphql_name='type')
    id = sgqlc.types.Field(String, graphql_name='id')
    geometry = sgqlc.types.Field(Any, graphql_name='geometry')
    properties = sgqlc.types.Field(Any, graphql_name='properties')


class FeatureCollection(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('features', 'type')
    features = sgqlc.types.Field(sgqlc.types.list_of(Feature), graphql_name='features')
    type = sgqlc.types.Field(sgqlc.types.non_null(GeoJSONObjectType), graphql_name='type')


class GeoBounds(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('north', 'south', 'east', 'west')
    north = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='north')
    south = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='south')
    east = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='east')
    west = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='west')


class HeadquartersMutation(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('update_calendar_event', 'delete_calendar_event', 'add_assignment_calendar_event', 'add_interview_calendar_event', 'delete_map', 'delete_user_from_map', 'add_user_to_map')
    update_calendar_event = sgqlc.types.Field(CalendarEvent, graphql_name='updateCalendarEvent', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('comment', sgqlc.types.Arg(String, graphql_name='comment', default=None)),
        ('new_start', sgqlc.types.Arg(sgqlc.types.non_null(DateTime), graphql_name='newStart', default=None)),
        ('public_key', sgqlc.types.Arg(sgqlc.types.non_null(Uuid), graphql_name='publicKey', default=None)),
        ('start_timezone', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='startTimezone', default=None)),
))
    )
    delete_calendar_event = sgqlc.types.Field(CalendarEvent, graphql_name='deleteCalendarEvent', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('public_key', sgqlc.types.Arg(sgqlc.types.non_null(Uuid), graphql_name='publicKey', default=None)),
))
    )
    add_assignment_calendar_event = sgqlc.types.Field(CalendarEvent, graphql_name='addAssignmentCalendarEvent', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('assignment_id', sgqlc.types.Arg(sgqlc.types.non_null(Int), graphql_name='assignmentId', default=None)),
        ('comment', sgqlc.types.Arg(String, graphql_name='comment', default=None)),
        ('new_start', sgqlc.types.Arg(sgqlc.types.non_null(DateTime), graphql_name='newStart', default=None)),
        ('start_timezone', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='startTimezone', default=None)),
))
    )
    add_interview_calendar_event = sgqlc.types.Field(CalendarEvent, graphql_name='addInterviewCalendarEvent', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('comment', sgqlc.types.Arg(String, graphql_name='comment', default=None)),
        ('interview_id', sgqlc.types.Arg(sgqlc.types.non_null(Uuid), graphql_name='interviewId', default=None)),
        ('new_start', sgqlc.types.Arg(sgqlc.types.non_null(DateTime), graphql_name='newStart', default=None)),
        ('start_timezone', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='startTimezone', default=None)),
))
    )
    delete_map = sgqlc.types.Field('Map', graphql_name='deleteMap', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('file_name', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='fileName', default=None)),
))
    )
    delete_user_from_map = sgqlc.types.Field('Map', graphql_name='deleteUserFromMap', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('file_name', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='fileName', default=None)),
        ('user_name', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='userName', default=None)),
))
    )
    add_user_to_map = sgqlc.types.Field('Map', graphql_name='addUserToMap', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('file_name', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='fileName', default=None)),
        ('user_name', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='userName', default=None)),
))
    )


class HeadquartersQuery(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('assignments', 'interviews', 'maps', 'questionnaires', 'questions', 'questionnaire_items', 'viewer', 'users', 'map_report')
    assignments = sgqlc.types.Field('IPagedConnectionOfAssignment', graphql_name='assignments', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('skip', sgqlc.types.Arg(Int, graphql_name='skip', default=None)),
        ('take', sgqlc.types.Arg(Int, graphql_name='take', default=None)),
        ('where', sgqlc.types.Arg(AssignmentsFilter, graphql_name='where', default=None)),
))
    )
    interviews = sgqlc.types.Field('IPagedConnectionOfInterview', graphql_name='interviews', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('skip', sgqlc.types.Arg(Int, graphql_name='skip', default=None)),
        ('take', sgqlc.types.Arg(Int, graphql_name='take', default=None)),
        ('where', sgqlc.types.Arg(InterviewsFilter, graphql_name='where', default=None)),
        ('order', sgqlc.types.Arg(sgqlc.types.list_of(sgqlc.types.non_null(InterviewSort)), graphql_name='order', default=None)),
))
    )
    maps = sgqlc.types.Field('IPagedConnectionOfMap', graphql_name='maps', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('skip', sgqlc.types.Arg(Int, graphql_name='skip', default=None)),
        ('take', sgqlc.types.Arg(Int, graphql_name='take', default=None)),
        ('where', sgqlc.types.Arg(MapsFilter, graphql_name='where', default=None)),
        ('order', sgqlc.types.Arg(sgqlc.types.list_of(sgqlc.types.non_null(MapsSort)), graphql_name='order', default=None)),
))
    )
    questionnaires = sgqlc.types.Field('IPagedConnectionOfQuestionnaire', graphql_name='questionnaires', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('skip', sgqlc.types.Arg(Int, graphql_name='skip', default=None)),
        ('take', sgqlc.types.Arg(Int, graphql_name='take', default=None)),
        ('id', sgqlc.types.Arg(Uuid, graphql_name='id', default=None)),
        ('version', sgqlc.types.Arg(Long, graphql_name='version', default=None)),
))
    )
    questions = sgqlc.types.Field(sgqlc.types.list_of(Entity), graphql_name='questions', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('id', sgqlc.types.Arg(sgqlc.types.non_null(Uuid), graphql_name='id', default=None)),
        ('version', sgqlc.types.Arg(sgqlc.types.non_null(Long), graphql_name='version', default=None)),
        ('language', sgqlc.types.Arg(String, graphql_name='language', default=None)),
        ('where', sgqlc.types.Arg(QuestionFilter, graphql_name='where', default=None)),
))
    )
    questionnaire_items = sgqlc.types.Field(sgqlc.types.list_of('QuestionnaireItem'), graphql_name='questionnaireItems', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('id', sgqlc.types.Arg(sgqlc.types.non_null(Uuid), graphql_name='id', default=None)),
        ('version', sgqlc.types.Arg(sgqlc.types.non_null(Long), graphql_name='version', default=None)),
        ('language', sgqlc.types.Arg(String, graphql_name='language', default=None)),
        ('where', sgqlc.types.Arg(QuestionnaireItemsFilter, graphql_name='where', default=None)),
))
    )
    viewer = sgqlc.types.Field('Viewer', graphql_name='viewer')
    users = sgqlc.types.Field('Users', graphql_name='users', args=sgqlc.types.ArgDict((
        ('skip', sgqlc.types.Arg(Int, graphql_name='skip', default=None)),
        ('take', sgqlc.types.Arg(Int, graphql_name='take', default=None)),
        ('order', sgqlc.types.Arg(sgqlc.types.list_of(sgqlc.types.non_null(UsersSortInput)), graphql_name='order', default=None)),
        ('where', sgqlc.types.Arg(UsersFilterInput, graphql_name='where', default=None)),
))
    )
    map_report = sgqlc.types.Field('MapReportHolder', graphql_name='mapReport', args=sgqlc.types.ArgDict((
        ('workspace', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='workspace', default='primary')),
        ('questionnaire_id', sgqlc.types.Arg(sgqlc.types.non_null(Uuid), graphql_name='questionnaireId', default=None)),
        ('questionnaire_version', sgqlc.types.Arg(Long, graphql_name='questionnaireVersion', default=None)),
        ('variable', sgqlc.types.Arg(String, graphql_name='variable', default=None)),
        ('zoom', sgqlc.types.Arg(sgqlc.types.non_null(Int), graphql_name='zoom', default=None)),
        ('client_map_width', sgqlc.types.Arg(sgqlc.types.non_null(Int), graphql_name='clientMapWidth', default=None)),
        ('east', sgqlc.types.Arg(sgqlc.types.non_null(Float), graphql_name='east', default=None)),
        ('west', sgqlc.types.Arg(sgqlc.types.non_null(Float), graphql_name='west', default=None)),
        ('north', sgqlc.types.Arg(sgqlc.types.non_null(Float), graphql_name='north', default=None)),
        ('south', sgqlc.types.Arg(sgqlc.types.non_null(Float), graphql_name='south', default=None)),
        ('where', sgqlc.types.Arg(MapReportFilter, graphql_name='where', default=None)),
))
    )


class IPagedConnectionOfAssignment(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('nodes', 'total_count', 'filtered_count')
    nodes = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(Assignment)), graphql_name='nodes')
    total_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='totalCount')
    filtered_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='filteredCount')


class IPagedConnectionOfInterview(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('nodes', 'total_count', 'filtered_count')
    nodes = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of('Interview')), graphql_name='nodes')
    total_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='totalCount')
    filtered_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='filteredCount')


class IPagedConnectionOfMap(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('nodes', 'total_count', 'filtered_count')
    nodes = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of('Map')), graphql_name='nodes')
    total_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='totalCount')
    filtered_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='filteredCount')


class IPagedConnectionOfQuestionnaire(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('nodes', 'total_count', 'filtered_count')
    nodes = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of('Questionnaire')), graphql_name='nodes')
    total_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='totalCount')
    filtered_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='filteredCount')


class IdentifyingEntity(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('answer_value', 'entity', 'value', 'value_bool', 'value_date', 'value_long', 'value_double', 'is_enabled')
    answer_value = sgqlc.types.Field(Int, graphql_name='answerValue')
    entity = sgqlc.types.Field(sgqlc.types.non_null(Entity), graphql_name='entity')
    value = sgqlc.types.Field(String, graphql_name='value')
    value_bool = sgqlc.types.Field(Boolean, graphql_name='valueBool')
    value_date = sgqlc.types.Field(DateTime, graphql_name='valueDate')
    value_long = sgqlc.types.Field(Long, graphql_name='valueLong')
    value_double = sgqlc.types.Field(Float, graphql_name='valueDouble')
    is_enabled = sgqlc.types.Field(Boolean, graphql_name='isEnabled')


class Interview(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('action_flags', 'assignment_id', 'id', 'status', 'interview_mode', 'responsible_name', 'responsible_id', 'responsible_role', 'supervisor_name', 'was_completed', 'created_date', 'key', 'client_key', 'update_date_utc', 'received_by_interviewer_at_utc', 'errors_count', 'questionnaire_id', 'questionnaire_variable', 'questionnaire_version', 'identifying_data', 'not_answered_count', 'calendar_event', 'cawi_link')
    action_flags = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(InterviewActionFlags))), graphql_name='actionFlags')
    assignment_id = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='assignmentId')
    id = sgqlc.types.Field(sgqlc.types.non_null(ID), graphql_name='id')
    status = sgqlc.types.Field(sgqlc.types.non_null(InterviewStatus), graphql_name='status')
    interview_mode = sgqlc.types.Field(sgqlc.types.non_null(InterviewMode), graphql_name='interviewMode')
    responsible_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='responsibleName')
    responsible_id = sgqlc.types.Field(sgqlc.types.non_null(Uuid), graphql_name='responsibleId')
    responsible_role = sgqlc.types.Field(sgqlc.types.non_null(UserRoles), graphql_name='responsibleRole')
    supervisor_name = sgqlc.types.Field(String, graphql_name='supervisorName')
    was_completed = sgqlc.types.Field(sgqlc.types.non_null(Boolean), graphql_name='wasCompleted')
    created_date = sgqlc.types.Field(sgqlc.types.non_null(DateTime), graphql_name='createdDate')
    key = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='key')
    client_key = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='clientKey')
    update_date_utc = sgqlc.types.Field(sgqlc.types.non_null(DateTime), graphql_name='updateDateUtc')
    received_by_interviewer_at_utc = sgqlc.types.Field(DateTime, graphql_name='receivedByInterviewerAtUtc')
    errors_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='errorsCount')
    questionnaire_id = sgqlc.types.Field(sgqlc.types.non_null(Uuid), graphql_name='questionnaireId')
    questionnaire_variable = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='questionnaireVariable')
    questionnaire_version = sgqlc.types.Field(sgqlc.types.non_null(Long), graphql_name='questionnaireVersion')
    identifying_data = sgqlc.types.Field(sgqlc.types.list_of(IdentifyingEntity), graphql_name='identifyingData')
    not_answered_count = sgqlc.types.Field(Int, graphql_name='notAnsweredCount')
    calendar_event = sgqlc.types.Field(CalendarEvent, graphql_name='calendarEvent')
    cawi_link = sgqlc.types.Field(String, graphql_name='cawiLink')


class Map(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('file_name', 'size', 'import_date', 'import_date_utc', 'users', 'x_max_val', 'y_max_val', 'x_min_val', 'y_min_val', 'wkid', 'max_scale', 'min_scale')
    file_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='fileName')
    size = sgqlc.types.Field(sgqlc.types.non_null(Long), graphql_name='size')
    import_date = sgqlc.types.Field(DateTime, graphql_name='importDate') # before 21.05
    import_date_utc = sgqlc.types.Field(DateTime, graphql_name='importDateUtc')
    users = sgqlc.types.Field(sgqlc.types.list_of('UserMap'), graphql_name='users')
    x_max_val = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='xMaxVal')
    y_max_val = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='yMaxVal')
    x_min_val = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='xMinVal')
    y_min_val = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='yMinVal')
    wkid = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='wkid')
    max_scale = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='maxScale')
    min_scale = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='minScale')


class MapReport(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('total_point', 'initial_bounds', 'feature_collection')
    total_point = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='totalPoint')
    initial_bounds = sgqlc.types.Field(GeoBounds, graphql_name='initialBounds')
    feature_collection = sgqlc.types.Field(FeatureCollection, graphql_name='featureCollection')


class MapReportHolder(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('report',)
    report = sgqlc.types.Field(sgqlc.types.non_null(MapReport), graphql_name='report')


class Questionnaire(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('variable', 'questionnaire_id', 'version', 'id', 'title', 'default_language_name', 'translations')
    variable = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='variable')
    questionnaire_id = sgqlc.types.Field(sgqlc.types.non_null(Uuid), graphql_name='questionnaireId')
    version = sgqlc.types.Field(sgqlc.types.non_null(Long), graphql_name='version')
    id = sgqlc.types.Field(ID, graphql_name='id')
    title = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='title')
    default_language_name = sgqlc.types.Field(String, graphql_name='defaultLanguageName')
    translations = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('Translation'))), graphql_name='translations')


class QuestionnaireItem(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('entity_type', 'title', 'variable', 'scope', 'label', 'type', 'variable_type', 'identifying', 'included_in_reporting_at_utc', 'options')
    entity_type = sgqlc.types.Field(EntityType, graphql_name='entityType')
    title = sgqlc.types.Field(String, graphql_name='title')
    variable = sgqlc.types.Field(String, graphql_name='variable')
    scope = sgqlc.types.Field(QuestionScope, graphql_name='scope')
    label = sgqlc.types.Field(String, graphql_name='label')
    type = sgqlc.types.Field(QuestionType, graphql_name='type')
    variable_type = sgqlc.types.Field(VariableType, graphql_name='variableType')
    identifying = sgqlc.types.Field(Boolean, graphql_name='identifying')
    included_in_reporting_at_utc = sgqlc.types.Field(DateTime, graphql_name='includedInReportingAtUtc')
    options = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(CategoricalOption))), graphql_name='options')


class Translation(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('id', 'name')
    id = sgqlc.types.Field(sgqlc.types.non_null(Uuid), graphql_name='id')
    name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='name')


class User(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('id', 'role', 'user_name', 'full_name', 'email', 'phone_number', 'creation_date', 'is_locked', 'is_archived', 'workspaces')
    id = sgqlc.types.Field(sgqlc.types.non_null(ID), graphql_name='id')
    role = sgqlc.types.Field(sgqlc.types.non_null(UserRoles), graphql_name='role')
    user_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='userName')
    full_name = sgqlc.types.Field(String, graphql_name='fullName')
    email = sgqlc.types.Field(String, graphql_name='email')
    phone_number = sgqlc.types.Field(String, graphql_name='phoneNumber')
    creation_date = sgqlc.types.Field(Date, graphql_name='creationDate')
    is_locked = sgqlc.types.Field(Boolean, graphql_name='isLocked')
    is_archived = sgqlc.types.Field(Boolean, graphql_name='isArchived')
    workspaces = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(String))), graphql_name='workspaces')


class UserMap(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('user_name',)
    user_name = sgqlc.types.Field(String, graphql_name='userName')


class Users(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('nodes', 'total_count', 'filtered_count')
    nodes = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(User)), graphql_name='nodes')
    total_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='totalCount')
    filtered_count = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='filteredCount')


class Viewer(sgqlc.types.Type):
    __schema__ = headquarters_schema
    __field_names__ = ('id', 'role', 'user_name', 'workspaces')
    id = sgqlc.types.Field(sgqlc.types.non_null(ID), graphql_name='id')
    role = sgqlc.types.Field(sgqlc.types.non_null(UserRoles), graphql_name='role')
    user_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='userName')
    workspaces = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(String))), graphql_name='workspaces')



########################################################################
# Unions
########################################################################

########################################################################
# Schema Entry Points
########################################################################
headquarters_schema.query_type = HeadquartersQuery
headquarters_schema.mutation_type = HeadquartersMutation
headquarters_schema.subscription_type = None




from typing import List

from .base import HQBase
from .exceptions import FeatureNotSupported
from .headquarters import Client
from .models import Version
from .utils import fix_qid


class WorkspacesApi(HQBase):
    """ Set of functions to access and manipulate Workspaces. """

    _apiprefix = "/api/v1/workspaces"

    def __init__(self, client: Client):
        if client.version < Version("21.01 (build 30293)"):
            raise FeatureNotSupported

        super().__init__(client)

    @fix_qid(expects={'user_id': 'string'})
    def get_list(self, user_id: str = None, include_disabled: bool = False):
        path = self.url
        length = 10
        start = 0
        total_count = 11
        params = {
            'start': start,
            'length': length,
            'IncludeDisabled': include_disabled,
            'UserId': user_id
        }
        while start < total_count:
            params['start'] = start
            r = self._make_call('get', path, params=params)
            if r.get("TotalCount"):
                total_count = r['TotalCount']
                yield from r['Workspaces']
            else:
                yield from ()
            start += length

    def get_info(self, name: str):
        path = self.url + '/{}'.format(name)
        return self._make_call('get', path)

    def create(self, name: str, display_name: str):
        path = self.url
        return self._make_call('post', path, json={'Name': name, 'DisplayName': display_name})

    def update(self, name: str, display_name: str):
        path = self.url + '/{}'.format(name)
        return self._make_call('patch', path, json={'DisplayName': display_name})

    def delete(self, name: str):
        path = self.url + '/{}'.format(name)
        _ = self._make_call('delete', path)

        return True

    def enable(self, name: str):
        path = self.url + '/{}/enable'.format(name)
        return self._make_call('post', path)

    def disable(self, name: str):
        path = self.url + '/{}/disable'.format(name)
        return self._make_call('post', path)

    def assign(self, user_ids: List[str], workspaces: List[str], supervisors: List[str] = None, mode: str = "add"):
        path = self.url + '/assign'

        if not isinstance(user_ids, list):
            user_ids = [user_ids]
        if not isinstance(workspaces, list):
            workspaces = [workspaces]
        if not isinstance(supervisors, list):
            supervisors = [supervisors]
        if mode.lower() not in ["add", "assign", "remove"]:
            raise ValueError("mode parameter must be one of 'add', 'assign', or 'remove'")

        data = {
            "UserIds": user_ids,
            "Workspaces": [{"workspace": w, "supervisorId": s} for w, s in zip(workspaces, supervisors)],
            "Mode": mode
        }
        return self._make_call('post', path, json=data)

    def status(self, name: str):
        path = self.url + '/status/{}'.format(name)
        return self._make_call('get', path)



from html import escape
from typing import Generator
from uuid import UUID

from sgqlc.operation import Operation

from .base import HQBase
from .headquarters_schema import (
    CalendarEvent,
    HeadquartersQuery,
    Interview,
    InterviewsFilter,
)
from .models import InterviewAnswers
from .utils import filter_object, fix_qid, order_object


class InterviewsApi(HQBase):
    """ Set of functions to access and manipulate Interviews. """

    _apiprefix = "/api/v1/interviews"

    @fix_qid(expects={'questionnaire_id': 'hex'})
    def get_list(self, fields: list = [], order=None,
                 skip: int = None, take: int = None, where: InterviewsFilter = None,
                 include_calendar_events: bool = False, **kwargs
                 ) -> Generator[Interview, None, None]:
        """Get list of interviews

        :param fields: list of fields to return, of ommited, all fields wll be returned
        :param order: list of fields to sort the results by
        :param skip: number of interviews to skip
        :param take: number of interviews to return
        :param questionnaire_id: filter by specific questionnaire id
        :param questionnaire_variable: filter by questionnaire variable instead
        :param questionnaire_version: filter by specific version number
        :include_calendar_events: return calendar_event object as part of the fields

        :returns: list of :class:`Interview` objects
        """
        interview_args = {
            "workspace": self.workspace
        }
        if order:
            interview_args["order"] = order_object("InterviewSort", order)
        if skip:
            interview_args["skip"] = skip
        if take:
            interview_args["take"] = take

        if where or kwargs:
            interview_args['where'] = filter_object("InterviewsFilter", where=where, **kwargs)

        if not fields:
            fields = [
                'id',
                'questionnaire_id',
                'questionnaire_version',
                'assignment_id',
                'responsible_id',
                'errors_count',
                'status',
            ]

        op = Operation(HeadquartersQuery)
        q = op.interviews(**interview_args)
        q.nodes.__fields__(*fields)
        if include_calendar_events:
            if type(include_calendar_events) in [list, tuple]:
                q.nodes.calendar_event.__fields__(*include_calendar_events)
            else:
                q.nodes.calendar_event.__fields__()
        cont = self._make_graphql_call(op)

        res = (op + cont).interviews

        yield from res.nodes

    def get_info(self, interview_id: UUID) -> list:
        path = self.url + '/{}'.format(interview_id)
        ret = self._make_call('get', path)
        if "Answers" in ret:
            obj = InterviewAnswers()
            obj.from_dict(ret["Answers"])
            return obj

    def delete(self, interviewid: UUID, comment: str = None):
        return self._change_status(action='delete', interviewid=interviewid, comment=comment)

    def get_calendar_event(self, interview_key: str) -> CalendarEvent:
        """Get calendar event associated with the interview
        """
        interview = self._get_interview_by_key(fields=["id"],
                                               key=interview_key,
                                               include_calendar_events=True)

        return interview.calendar_event

    def set_calendar_event(self, interview_key: str, new_start: str,
                           start_timezone: str, comment: str = None) -> CalendarEvent:
        """Add new calendar event to the interview, or update the existing one

        :param interview_key: key of the interview to update
        :param new_start: start date of the calendar event
        :param start_timezone: timezone string for the start date
        :param comment: add comment to the calendar event

        :returns: :class:`CalendarEvent' object
        """
        interview = self._get_interview_by_key(fields=["id"],
                                               key=interview_key,
                                               include_calendar_events=["public_key"])

        kwargs = {
            "new_start": new_start,
            "start_timezone": start_timezone,
            "comment": comment
        }
        if hasattr(interview.calendar_event, "public_key"):
            kwargs["method_name"] = "update_calendar_event"
            kwargs["public_key"] = interview.calendar_event.public_key
        else:
            kwargs["method_name"] = "add_interview_calendar_event"
            kwargs["interview_id"] = interview.id

        return self._call_mutation(**kwargs)

    def delete_calendar_event(self, interview_key: str) -> CalendarEvent:
        """Remove calendar event associated with the interview
        """
        interview = self._get_interview_by_key(fields=["id"],
                                               key=interview_key,
                                               include_calendar_events=["public_key"])

        if hasattr(interview.calendar_event, "public_key"):
            return self._call_mutation(method_name="delete_calendar_event",
                                       public_key=interview.calendar_event.public_key)

    def approve(self, interviewid, comment=''):
        return self._change_status(action='approve', interviewid=interviewid, comment=comment)

    def assign(self, interviewid, responsibleid, responsiblename=''):
        return self._reassign(action='assign', interviewid=interviewid,
                              responsibleid=responsibleid, responsiblename=responsiblename)

    def assign_supervisor(self, interviewid, responsibleid, responsiblename=''):
        return self._reassign(action='assignsupervisor', interviewid=interviewid,
                              responsibleid=responsibleid, responsiblename=responsiblename)

    @fix_qid(expects={'interview_id': 'string', 'question_id': 'hex'})
    def comment(self, interview_id, comment, question_id: str = None, variable: str = None, roster_vector: list = []):
        params = {'comment': escape(comment)}
        if roster_vector:
            params['rosterVector'] = roster_vector

        if variable:
            path = self.url + '/{}/comment-by-variable/{}'.format(interview_id, variable)
        else:
            if question_id:
                path = self.url + '/{}/comment/{}'.format(interview_id, question_id)
            else:
                raise TypeError("comment() either 'variable' or 'question_id' argument is required")

        self._make_call('post', path, params=params)

    @fix_qid(expects={'interview_id': 'string'})
    def history(self, interview_id):
        path = self.url + '/{}/history'.format(interview_id)
        ret = self._make_call('get', path)
        yield from ret["Records"]

    def hqapprove(self, interviewid, comment=''):
        return self._change_status(action='hqapprove', interviewid=interviewid, comment=comment)

    def hqreject(self, interviewid, comment=''):
        return self._change_status(action='hqreject', interviewid=interviewid, comment=comment)

    def hqunapprove(self, interviewid, comment=''):
        return self._change_status(action='hqunapprove', interviewid=interviewid, comment=comment)

    def reject(self, interviewid, comment=''):
        return self._change_status(action='reject', interviewid=interviewid, comment=comment)

    def _change_status(self, interviewid, action, comment=None):
        path = self.url + '/{}/{}'.format(interviewid, action)
        params = {'comment': escape(comment)} if comment else {}
        return self._make_call('patch', path, params=params)

    def _reassign(self, interviewid, action, responsibleid, responsiblename=''):
        path = self.url + '/{}/{}'.format(interviewid, action)
        payload = {
            'ResponsibleId': responsibleid,
            'ResponsibleName': responsiblename
        }
        _ = self._make_call('patch', path, json=payload)
        return True

    def _get_interview_by_key(self, **kwargs):
        # expecting fields, key, include_calendar_events
        try:
            interview = next(self.get_list(**kwargs))
        except StopIteration:
            raise ValueError("interview was not found")

        return interview



import uuid
from datetime import datetime
from functools import wraps

from .exceptions import IncompleteQuestionnaireIdError
from .headquarters_schema import headquarters_schema as schema


def fix_qid(expects: dict = {'questionnaire_id': 'hex'}):
    def wrapper_outer(func):
        @wraps(func)
        def wrapper_inner(*args, **kwargs):
            for param_name in expects.keys():
                if param_name in kwargs:
                    if expects[param_name] == 'hex':
                        kwargs[param_name] = uuid.UUID(kwargs[param_name]).hex
                    elif expects[param_name] == 'string':
                        try:
                            kwargs[param_name] = str(uuid.UUID(kwargs[param_name]))
                        except (AttributeError, ValueError):
                            raise ValueError(f"{param_name} expect a valid uuid string")
                    else:
                        raise ValueError('expects should be either hex or string')
            return func(*args, **kwargs)
        return wrapper_inner
    return wrapper_outer


def to_hex(q_id):
    return uuid.UUID(str(q_id)).hex


def to_qidentity(q_id, q_version):
    return "{}${}".format(to_hex(q_id), q_version)


def parse_qidentity(q_identity):
    if type(q_identity) is tuple:
        (q_id, q_version) = q_identity
    else:
        qq = q_identity.split("$")
        try:
            q_id = qq[0]
            q_version = qq[1]
        except IndexError:
            raise(IncompleteQuestionnaireIdError)

    return to_qidentity(q_id, q_version)


def to_camel(string: str) -> str:
    return ''.join(word.capitalize() for word in string.split('_'))


def get_properties(obj, types: list = [], properties: list = [], groups: bool = False, items: bool = True) -> dict:
    ret = {}
    if type(obj).__name__ in ["Group", "QuestionnaireDocument"]:
        if groups:
            ret[obj.public_key.hex] = obj
        for ch in obj.children:
            ret.update(get_properties(ch, types, properties, groups, items))
    elif items:
        type_name = type(obj).__name__
        if type_name in types or not types:
            if properties:
                ret[obj.variable_name] = {p: getattr(obj, p) for p in properties if hasattr(obj, p)}
            else:
                ret[obj.variable_name] = obj
    return ret


def order_object(classname: str, params):
    if type(params) in [list, tuple]:
        d = {}
        for item in params:
            if type(item) is tuple:
                d[item[0]] = item[1]
            elif type(item) is str:
                d[item] = "ASC"
            else:
                raise TypeError("Elements must be either string or tuple")
        params = d
    elif type(params) is not dict:
        raise TypeError("Argument must be dict, list, or tuple")

    return [getattr(schema, classname)(**{item[0]: item[1]}) for item in params.items()]


def filter_object(classname: str, where=None, **kwargs):
    filter_type = getattr(schema, classname)
    if where and type(where) is not filter_type:
        raise TypeError(f"where parameter must be an object of type {classname}")

    fields = [item for item in getattr(filter_type, "__field_names__") if item not in ["and_", "or_"]]
    filter_args = {}
    for key, value in kwargs.items():
        if key not in fields:
            raise KeyError(f"{classname} does not contain field {key}")
        field_type = getattr(getattr(schema, classname), key).type
        if key in ["responsible_name", "supervisor_name"]:
            value = value.lower()
        filter_args[key] = field_type(eq=value)

    if where:
        if filter_args:
            return filter_type(and_=[where, filter_type(**filter_args)])
        else:
            return where
    else:
        if filter_args:
            return filter_type(**filter_args)


def parse_date(date_string: str) -> datetime:
    try:
        return datetime.strptime(f"{date_string}+0000", "%Y-%m-%dT%H:%M:%S.%f%z")
    except (TypeError, ValueError):
        return date_string



from datetime import datetime
from typing import Generator
from uuid import UUID

from sgqlc.operation import Operation

from .base import HQBase
from .headquarters_schema import HeadquartersQuery, UsersFilterInput
from .headquarters_schema import User as GraphQLUser, Viewer
from .models import InterviewerAction, User
from .utils import filter_object, order_object


class UsersApi(HQBase):
    """ Set of functions to access and manipulate Users. """

    _apiprefix = "/api/v1"

    def get_list(self, fields: list = [],
                 order=None, skip: int = None, take: int = None,
                 where: UsersFilterInput = None, **kwargs) -> Generator[GraphQLUser, None, None]:

        q_args = {
        }
        if order:
            q_args["order"] = order_object("UsersSortInput", order)
        if skip:
            q_args["skip"] = skip
        if take:
            q_args["take"] = take
        if where or kwargs:
            q_args['where'] = filter_object("UsersFilterInput", where=where, **kwargs)

        op = Operation(HeadquartersQuery)
        q = op.users(**q_args)
        q.nodes.__fields__(*fields)
        cont = self._make_graphql_call(op)
        res = (op + cont).users

        yield from res.nodes

    def get_info(self, id):
        path = self._url_users + '/{}'.format(id)
        return self._make_call('get', path)

    def get_actions_log(
            self, id: UUID, start: datetime = None, end: datetime = None) -> Generator[InterviewerAction, None, None]:

        path = self._url_interviewers + '/{}/actions-log'.format(id)
        params = {}
        if start:
            params["start"] = start.strftime("%Y-%m-%d")
        if end:
            params["end"] = end.strftime("%Y-%m-%d")
        response = self._make_call('get', path, params=params)
        for ac in response:
            yield InterviewerAction(**ac)

    def list_supervisors(self):
        path = self._url_supervisors
        return self._list_users(path)

    def list_interviewers(self, id):
        path = self._url_supervisors + '/{}/interviewers'.format(id)
        return self._list_users(path)

    def unarchive(self, id):
        path = self._url_users + '/{}/unarchive'.format(id)
        return self._make_call('patch', path)

    def archive(self, id):
        path = self._url_users + '/{}/archive'.format(id)
        return self._make_call('patch', path)

    def create(self, **kwargs):

        user = User(**kwargs)
        path = self._url_users

        return self._make_call("post", path, data=user.json(by_alias=True),
                               headers={"content-type": "application/json"})

    def viewer(self, username: str = None, password: str = None) -> Viewer:
        op = Operation(HeadquartersQuery)
        op.viewer()
        if username and password:
            cont = self._make_graphql_call(op, auth=(username, password))
        else:
            cont = self._make_graphql_call(op)

        return (op + cont).viewer

    def _list_users(self, path):
        page_size = 10
        page = 1
        total_count = 11
        params = {
            'offset': page,
            'limit': page_size
        }
        while (page - 1) * page_size < total_count:
            params['offset'] = page
            r = self._make_call('get', path, params=params)
            total_count = r['TotalCount']
            if total_count > 0:
                yield from r['Users']
            else:
                yield from ()
            page += 1

    @property
    def _url_users(self):
        return self.url + '/users'

    @property
    def _url_supervisors(self):
        return self.url + '/supervisors'

    @property
    def _url_interviewers(self):
        return self.url + '/interviewers'



from .__about__ import __version__
from .assignments import AssignmentsApi
from .export import ExportApi
from .headquarters import Client
from .interviews import InterviewsApi
from .maps import MapsApi
from .questionnaires import QuestionnairesApi
from .settings import SettingsApi
from .users import UsersApi
from .workspaces import WorkspacesApi


__all__ = ["__version__", "AssignmentsApi", "ExportApi",
           "Client", "InterviewsApi", "MapsApi", "QuestionnairesApi",
           "SettingsApi", "UsersApi", "WorkspacesApi", ]



from typing import Generator

from sgqlc.operation import Operation

from .base import HQBase
from .exceptions import NotAcceptableError
from .headquarters_schema import (
    HeadquartersQuery,
    ListFilterInputTypeOfUserMapFilterInput,
    Map,
    MapsFilter,
    StringOperationFilterInput,
    UserMapFilterInput
)
from .models import Version
from .utils import order_object


class MapsApi(HQBase):
    """ Set of functions to access and manipulate Maps. """

    def get_list(self, filter_user: str = None, fields: list = [],
                 order: list = None, skip: int = None, take: int = None, **kwargs) -> Generator[Map, None, None]:
        """Get list of maps

        :param filter_user: List only maps linked to the user

        :param fields: List of fields to return for each map

        :returns: List of Map objects
        """

        if filter_user:
            kwargs["users"] = ListFilterInputTypeOfUserMapFilterInput(
                some=UserMapFilterInput(
                    user_name=StringOperationFilterInput(
                        eq=filter_user)))
        maps_args = {
            "workspace": self.workspace
        }
        if kwargs:
            maps_args["where"] = MapsFilter(**kwargs)
        if order:
            maps_args["order"] = order_object("MapsSort", order)
        if skip:
            maps_args["skip"] = skip
        if take:
            maps_args["take"] = take

        if not fields:
            fields = self._default_fields()

        op = Operation(HeadquartersQuery)
        q = op.maps(**maps_args)
        q.__fields__('filtered_count')
        q.nodes.__fields__(*fields)

        cont = self._make_graphql_call(op)
        res = (op + cont).maps

        yield from res.nodes

    def delete(self, file_name: str) -> Map:
        """Delete a map file

        :param file_name: Filename (with extension) to be deleted
        """
        return self._call_mutation(method_name="delete_map",
                                   file_name=file_name,
                                   fields=self._default_fields())

    def upload(self, zip_file) -> bool:
        """Upload a zip file with maps
        Currently only works for the admin user!

        :param file_name: Archive with extension with the maps

        :returns: `True` if successful, otherwise raises `NotAcceptableError`
        """
        ret = self._make_call(method="post", path=f"{self.url}/api/MapsApi/Upload",
                              files={'file': open(zip_file, 'rb')}, use_login_session=True)
        if ret["isSuccess"]:
            return True
        else:
            raise NotAcceptableError(ret["errors"][0])

    def add_user(self, file_name: str, user_name: str) -> Map:
        """Add user-to-map link

        :param file_name: Filename of the map
        :param user_name: Interviewer role user name

        :returns: Modified Map object
        """
        return self._call_mutation(method_name="add_user_to_map",
                                   file_name=file_name, user_name=user_name,
                                   fields=self._default_fields())

    def delete_user(self, file_name: str, user_name: str) -> Map:
        """Remove user-to-map link

        :param file_name: Filename of the map
        :param user_name: Interviewer role user name

        :returns: Modified Map object
        """
        return self._call_mutation(method_name="delete_user_from_map",
                                   file_name=file_name,
                                   user_name=user_name,
                                   fields=self._default_fields())

    def _default_fields(self):
        if self._hq.version < Version("21.05 (build 31160)"):
            return [
                "file_name",
                "import_date",
            ]
        else:
            return [
                "file_name",
                "import_date_utc",
            ]



import time
from datetime import datetime, timedelta, timezone

from .base import HQBase
from .models import ExportJob
from .utils import parse_qidentity


class ExportApi(HQBase):
    """ Set of functions to access and generate export files. """

    _apiprefix: str = "/api/v2/export"

    def get_list(self, export_type=None, interview_status=None,
                 questionnaire_identity=None, export_status=None,
                 has_file=None):
        path = self.url
        params = {
            "exportType": export_type,
            "interviewStatus": interview_status,
            "questionnaireIdentity": parse_qidentity(questionnaire_identity),
            "exportStatus": export_status,
            "hasFile": has_file,
        }
        r = self._make_call('get', path, params=params)
        for item in r:
            yield ExportJob.from_dict(item)

    def get(self, questionnaire_identity: str,
            export_type: str = "Tabular", interview_status="All",
            export_path: str = "",
            generate: bool = False, limit_age: int = None, limit_date: datetime = None, show_progress: bool = False):
        """Downloads latest available export file

        :param questionnaire_identity: Questionnaire id in format QuestionnaireGuid$Version

        :param export_type: Format of the export data:
            ``Tabular``, ``STATA``, ``SPSS``, ``Binary``, ``DDI``, ``Paradata``

        :param interview_status: What interviews to include in the export:
            ``All``, ``SupervisorAssigned``, ``InterviewerAssigned``, ``Completed``,
            ``RejectedBySupervisor``, ``ApprovedBySupervisor``, ``RejectedByHeadquarters``, ``ApprovedByHeadquarters``

        :param export_path: Path to save the downloaded file
        :param generate: generate new export if no existing result setisfies the specified filters
        :param limit_age: only return export file if created less than ``limit_age`` minutes ago
        :param limit_date: only return export file if created after the limit_date
        """

        qid = parse_qidentity(questionnaire_identity)
        common_args = {
            "questionnaire_identity": qid,
            "export_type": export_type,
            "interview_status": interview_status,
        }

        try:
            job = self._get_first_suitable(common_args, limit_age, limit_date)
            if job:
                return self._make_call('get', job.download_link, filepath=export_path, stream=True)
        except StopIteration:
            pass

        if not generate:
            return

        job = self.start(ExportJob(**common_args), wait=True, show_progress=show_progress)
        if job.has_export_file:
            response = self._make_call('get', job.download_link, filepath=export_path, stream=True)

        if show_progress:
            print(f"Archive was downloaded to {response}")

        return response

    def _get_first_suitable(self, common_args, limit_age=None, limit_date=None):
        ret_list = self.get_list(**common_args, export_status="Completed", has_file="true")

        if limit_date is None:
            limit_date = datetime(1970, 1, 1)

        if limit_date.tzinfo is None:
            limit_date = limit_date.astimezone()  # interpret date as in local timezone

        if limit_age:
            limit1 = datetime.now().astimezone() - timedelta(minutes=limit_age)
            if limit1 > limit_date:
                limit_date = limit1

        # start_date is returned in UTC, convert creation_limit as well
        limit_date = limit_date.astimezone(timezone(timedelta(0)))
        for job in ret_list:
            if job.start_date > limit_date:
                return job

    def get_info(self, job_id: int) -> ExportJob:
        path = self.url + '/{}'.format(job_id)
        return ExportJob.from_dict(self._make_call('get', path))

    def start(self, export_job: ExportJob, wait: bool = False, show_progress: bool = False) -> ExportJob:
        """Start new export job

        :param export_job: ``ExportJob`` object
        :param wait: if ``True`` will wait for the process to complete, otherwise, exit right away

        :returns: ``ExportJob`` object
        """
        path = self.url
        job = ExportJob.from_dict(self._make_call("post", path, json=export_job.to_json()))
        if wait:
            job = self.get_info(job.job_id)

            if show_progress:
                print("Generating...")

            while job.export_status != "Completed":
                time.sleep(1)
                if show_progress:
                    print(".", end="", flush=True)
                job = self.get_info(job.job_id)
            if show_progress:
                print()

        return job

    def cancel(self, job_id: int) -> None:
        response = self.get_info(job_id)
        if response.export_status == "Running":
            path = self.url + '{}'.format(job_id)
            _ = self._make_call('delete', path)
        else:
            print('No running export process was found')



import csv
import zipfile
from io import TextIOWrapper
from tempfile import TemporaryDirectory
from uuid import UUID

from sgqlc.operation import Operation

from .base import HQBase
from .headquarters_schema import HeadquartersQuery
from .interviews import InterviewsApi
from .models import AssignmentWebLink, QuestionnaireDocument


class QuestionnairesApi(HQBase):
    """ Set of functions to access information on Questionnaires. """

    _apiprefix = "/api/v1/questionnaires"

    def get_list(self, fields: list = [], questionnaire_id: str = None, version: int = None,
                 skip: int = None, take: int = None):
        if not fields:
            fields = [
                "id",
                "questionnaire_id",
                "version",
                "title",
                "variable",
            ]
        # we always have workspace parameter
        q_args = {
            "workspace": self.workspace
        }
        if questionnaire_id:
            q_args["id"] = questionnaire_id
        if version:
            q_args["version"] = version
        if skip:
            q_args["skip"] = skip
        if take:
            q_args["take"] = take

        op = Operation(HeadquartersQuery)
        q = op.questionnaires(**q_args)
        q.nodes.__fields__(*fields)

        cont = self._make_graphql_call(op)

        res = (op + cont).questionnaires

        yield from res.nodes

    def statuses(self):
        path = self.url + '/statuses'
        return self._make_call('get', path)

    def document(self, id: UUID, version: int) -> QuestionnaireDocument:
        path = self.url + '/{}/{}/document'.format(id, version)
        return self._make_call('get', path, parser=QuestionnaireDocument.parse_raw)

    def interviews(self, id: UUID, version: int):
        api = InterviewsApi(client=self._hq)
        return api.get_list(questionnaire_id=id, questionnaire_version=version)

    def update_recordaudio(self, id: UUID, version: int, enabled: bool):
        path = self.url + '/{}/{}/recordAudio'.format(id, version)
        return self._make_call('post', path, json={"Enabled": enabled})

    def download_web_links(self, id: UUID, version: int, path: str = None):
        """Download links for the assignments in Web Mode.

        :param id: questionnaire id
        :param version: questionnaire version
        :param path: optionally specify the download location

        if `path` is specified, zip archive will be downloaded to the location.
        Otherwise, list of ``AssignmentWebLink`` objects will be returned
        """
        common_args = {
            "method": "get",
            "path": f"{self._hq.baseurl}/{self.workspace}/api/LinksExport/Download/{id}${version}",
            "stream": True,
            "use_login_session": True,
        }
        if path:
            return self._make_call(**common_args, filepath=path)

        with TemporaryDirectory() as tempdir:
            outfile = self._make_call(**common_args, filepath=tempdir)
            with zipfile.ZipFile(outfile, "r") as zip_ref:
                with zip_ref.open("interviews.tab") as infile:
                    data = csv.DictReader(TextIOWrapper(infile, 'utf-8'), delimiter="\t")
                    return [AssignmentWebLink(**row) for row in data]
